# -*- coding: utf-8 -*-
"""
.. Authors:
    Novimir Pablant <npablant@pppl.gov>

Define the :class:`InteractMirror` class.
"""

import numpy as np
from copy import deepcopy


from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.optics._InteractObject import InteractObject
from xicsrt.tools import xicsrt_math as xm

@dochelper
class InteractMirror(InteractObject):
    """
    A perfect mirror interaction.
    """

    def interact(self, rays, xloc, norm, mask=None):
        rays = self.reflect_vectors(rays, xloc, norm, mask)
        if mask is not None:
            rays['mask'][:] = mask
        return rays

    def reflect_vectors(self, rays, xloc, normals, mask=None):
        if mask is None: mask = rays['mask']

        O = rays['origin']
        D = rays['direction']
        m = mask

        # Perform reflection around normal vector, creating updating rays with
        # new origin O = X and new direction D
        O[:] = xloc[:]
        D[m] -= 2 * (np.einsum('ij,ij->i', D[m], normals[m])[:, np.newaxis]
                     * normals[m])

        return rays



# -*- coding: utf-8 -*-
"""
..  Authors:
    Novimir Antoniuk Pablant <npablant@pppl.gov>

An example showing how to define a complex aperture.
"""

import numpy as np
import xicsrt
xicsrt.warn_version('0.8')

config = {}

config['general'] = {}
config['general']['number_of_iter'] = 5
config['general']['save_images'] = False
config['general']['random_seed'] = 0

config['sources'] = {}
config['sources']['source'] = {}
config['sources']['source']['class_name'] = 'XicsrtSourceDirected'
config['sources']['source']['intensity'] = 1e3
config['sources']['source']['wavelength'] = 3.9492
config['sources']['source']['angular_dist'] = 'isotropic_xy'
config['sources']['source']['spread'] = np.radians(6.0)

config['optics'] = {}
config['optics']['aperture'] = {}
config['optics']['aperture']['class_name'] = 'XicsrtOpticAperture'
config['optics']['aperture']['origin'] = [0.0, 0.0, 0.8]
config['optics']['aperture']['zaxis'] = [0.0, 0.0, -1]
config['optics']['aperture']['aperture']=[
    {'shape':'circle', 'size':[0.075], 'logic':'and'},
    {'shape':'circle', 'size':[0.065], 'origin':[-0.010, -0.01],  'logic':'not'},
    {'shape':'circle', 'size':[0.048], 'origin':[-0.027, -0.01],  'logic':'or'},
    {'shape':'circle', 'size':[0.044], 'origin':[-0.032, -0.015], 'logic':'not'},
    {'shape':'circle', 'size':[0.034], 'origin':[-0.041, -0.013], 'logic':'or'},
    {'shape':'circle', 'size':[0.032], 'origin':[-0.045, -0.018], 'logic':'not'},
    {'shape':'circle', 'size':[0.025], 'origin':[-0.038, -0.020], 'logic':'or'},
    ]

config['optics']['detector'] = {}
config['optics']['detector']['class_name'] = 'XicsrtOpticDetector'
config['optics']['detector']['origin'] = [0.0, 0.0, 1.0]
config['optics']['detector']['zaxis'] = [0.0, 0.0, -1]
config['optics']['detector']['xsize'] = 0.2
config['optics']['detector']['ysize'] = 0.2


results = xicsrt.raytrace(config)


# -*- coding:utf-8 -*-
"""
.. Authors
    Novimir pablant <npablant@pppl.gov>
    Yevgeniy Yakusevich <eugenethree@gmail.com>
    James Kring <jdk0026@tigermail.auburn.edu>

A set of tools for 2d visualization of the XICSRT results
"""
import numpy as np
import logging

from matplotlib import pyplot

from xicsrt.util import mirplot
from xicsrt import xicsrt_public
from xicsrt.tools import xicsrt_aperture

from xicsrt.visual import detview

log = logging.getLogger(__name__)

def plot_example(results, name):
    """
    A simplified plotting routine to serve as an example of how to develop
    xicsrt visualizations. This function will plot found ray intersections.
    """

    # Retrieve an object for the given optic/source.
    obj = xicsrt_public.get_element(results['config'], name)

    # Transform from global coordinate to local optic coordinates
    origin_ext = results['found']['history'][name]['origin']
    origin_loc = obj.point_to_local(origin_ext)

    # Use the mirplot utility to generate the plot.
    plotlist = [{
        'type':'scatter',
        'x':origin_loc[:, 0],
        'y':origin_loc[:, 1],
        }]
    fig = mirplot.plot_to_screen(plotlist)

    return fig


def plot_intersect(*args, **kwargs):
    """
    Plot the intersection of rays with the given optic.

    Parameters
    ----------
    results
      The restults dictionary from `raytrace()` that include the ray history.

    Keywords
    --------
    name :string (None)
      The name of the optic or source for which to plot intersections. The name
      refers to the key of the entry in the config dictionary. For example
      the name 'detector' will refer to config['optics']['detector'].

    section :string (None)
      [Optional] The name of the config section in which to search for `name`.
      This should typically be either 'optics' or 'sources'. If no section is
      given then then 'optics' will be searched first, then 'sources'.

    options :dict (None)
      [Optional] A dictionary containing plot options. All options can also be
      passed individually as keywords.

    Returns
    -------
    plotlist
      Will return a plotlist with the full plot definition.
    """

    # Aspect is handled here, so turn it off in the plotlist.
    plotlist = _get_intersect_plotlist(*args, **kwargs, _noaspect=True)

    gs = {
        'width_ratios':[3, 1],
        'height_ratios':[1, 3],
        'wspace':0.1,
        'hspace':0.1,
    }
    fig, axs = pyplot.subplots(2, 2, gridspec_kw=gs)
    axs[0, 1].set_axis_off()

    axesdict = {
        'scatter':axs[1, 0],
        'xhist':axs[0, 0],
        'yhist':axs[1, 1],
    }

    for ax in axesdict.values():
        ax.label_outer()

    axesdict['xhist'].sharex(axesdict['scatter'])
    axesdict['yhist'].sharey(axesdict['scatter'])

    mirplot.plot_to_axes(plotlist, axesdict)

    if kwargs.get('aspect','auto') == 'equal':
        _update_lim_aspect(axesdict['scatter'])
        axesdict['scatter'].callbacks.connect('ylim_changed', _on_ylims_change)

    return fig


def _get_intersect_plotlist(
        results,
        name=None,
        section=None,
        options=None,
        _noaspect=False,
        **kwargs,
        ):
    """
    Return a plotlist for :func:`plot_intersect`.
    """

    if options is None:
        opt = {}
    else:
        opt = options
    opt.update(kwargs)

    opt.setdefault('name', name)
    opt.setdefault('found', True)
    opt.setdefault('lost', True)
    opt.setdefault('bounds', True)
    opt.setdefault('aperture', True)
    opt.setdefault('aspect', 'equal')
    opt.setdefault('scale', None)
    opt.setdefault('units', None)
    opt.setdefault('lost_color', 'royalblue')
    opt.setdefault('found_color', 'red')
    opt.setdefault('alpha', None)
    opt.setdefault('lost_alpha', None)
    opt.setdefault('found_alpha', None)
    opt.setdefault('xbound', None)
    opt.setdefault('ybound', None)
    opt.setdefault('hist_size', None)
    opt.setdefault('hist_bins', None)
    opt.setdefault('hist_norm', False)
    opt.setdefault('drawstyle', 'steps-mid')
    opt.setdefault('linewidth', None)
    opt.setdefault('found_max', None)
    opt.setdefault('lost_max', None)
    opt.setdefault('found_mask', None)
    opt.setdefault('lost_mask', None)

    if opt['name'] is None:
        opt['name'] = 'detector'

    if (opt['alpha'] is None) and (opt['lost_alpha'] is None):
        opt['lost_alpha'] = 0.1
    elif (opt['lost_alpha'] is None):
        opt['lost_alpha'] = opt['alpha']

    if (opt['alpha'] is None) and (opt['found_alpha'] is None):
        opt['found_alpha'] = 0.5
    elif (opt['found_alpha'] is None):
        opt['found_alpha'] = opt['alpha']

    if opt['scale'] is None:opt['scale'] = 1.0

    if not opt.get('xlabel'):
        opt['xlabel'] = f'x'
        if opt.get('units'):
            opt['xlabel'] += f" [{opt['units']}]"

    if not opt.get('ylabel'):
        opt['ylabel'] = f'y'
        if opt.get('units'):
            opt['ylabel'] += f" [{opt['units']}]"

    if _noaspect:
        opt['aspect'] = None

    plotlist = []

    config = results['config']
    name = opt['name']

    # Get the crystal object from the dispatcher.
    obj = xicsrt_public.get_element(config, name, section)

    if opt['xbound'] is None:
        opt['xbound'] = np.array([-1 * obj.param['xsize'] / 2, obj.param['xsize'] / 2]) * 1.2
        opt['xbound'] *= opt['scale']
    else:
        opt['xbound'] = np.asarray(opt['xbound'])*opt['scale']

    if opt['ybound'] is None:
        opt['ybound'] = np.array([-1 * obj.param['ysize'] / 2, obj.param['ysize'] / 2]) * 1.2
        opt['ybound'] *= opt['scale']
    else:
        opt['ybound'] = np.asarray(opt['ybound'])*opt['scale']

    plotlist.extend([{'type':'figure', 'suptitle':name}])

    # Begin building plotlist.
    if opt['bounds']:
        plotlist.extend(_get_bounds_plotlist(obj, opt['scale']))

    if opt['aperture']:
        plotlist.extend(_get_aperture_plotlist(obj, opt['scale']))

    plotlist.extend(_get_hist(obj, results, opt, raytype='found', axis=0))
    plotlist.extend(_get_hist(obj, results, opt, raytype='found', axis=1))

    if opt['lost']:
        plotlist.extend(_get_rays_plotlist(obj, results, opt, raytype='lost'))

    if opt['found']:
        plotlist.extend(_get_rays_plotlist(obj, results, opt, raytype='found'))

    return plotlist


def _get_hist(obj, results, opt, raytype='found', axis=0):
    prefix = raytype + '_'
    name = opt['name']

    origin_ext = results[raytype]['history'][name]['origin']
    origin_loc = obj.point_to_local(origin_ext)
    mask = np.all(~ np.isnan(origin_ext[:, :]), axis=1)

    mask_user = opt[prefix+'mask']
    if mask_user is not None:
        mask &= mask_user

    if axis == 0:
        range_ = opt['xbound']
        name = 'xhist'
    else:
        range_ = opt['ybound']
        name = 'yhist'

    if opt.get('hist_size'):
        binsize = opt['hist_size'] * opt['scale']
        bins = int(np.round((range_[1] - range_[0]) / binsize))
    elif opt.get('hist_bins'):
        bins = opt['hist_bins']
    elif obj.param.get('pixel_size'):
        binsize = obj.param['pixel_size'] * opt['scale']
        bins = int(np.round((range_[1] - range_[0]) / binsize))
    else:
        bins = 100

    # Some logging
    binsize = (range_[1] - range_[0])/bins
    log.info(f'Histogram bins, size:{binsize:0.6f} num:{bins:4d}')

    # Calculate the histogram.
    hist, bins = np.histogram(
        origin_loc[mask, axis] * opt['scale'],
        bins,
        range=range_)
    bins_c = (bins[0:-1] + bins[1:]) / 2

    if opt['hist_norm']:
        norm_scale = 1.0 / np.max(hist)
    else:
        norm_scale = 1.0

    if axis == 0:
        x = bins_c
        y = hist * norm_scale
    else:
        y = bins_c
        x = hist * norm_scale

    plotlist = []
    plotlist.append({
        'axes':name,
        'x':x,
        'y':y,
        'drawstyle':opt['drawstyle'],
        'color':'black',
        'linewidth':opt['linewidth'],
    })

    return plotlist


def _get_rays_plotlist(obj, results, opt, raytype='found'):
    prefix = raytype + '_'
    name = opt['name']

    origin_ext = results[raytype]['history'][name]['origin']
    origin_loc = obj.point_to_local(origin_ext)
    mask = np.all(~ np.isnan(origin_ext[:, :]), axis=1)

    mask_user = opt[prefix+'mask']
    if mask_user is not None:
        mask &= mask_user

    mask = _truncate_mask(mask, opt[prefix+'max'])

    log.debug(f'{raytype:5.5s} {name:10.10s} {sum(mask)}')

    plotlist = []
    if np.sum(mask) > 0:
        plotlist.append({
            'axes':'scatter',
            'type':'scatter',
            'x':origin_loc[mask, 0] * opt['scale'],
            'y':origin_loc[mask, 1] * opt['scale'],
            'xbound':opt['xbound'],
            'ybound':opt['ybound'],
            'aspect':opt['aspect'],
            'alpha':opt[raytype + '_alpha'],
            'color':opt[raytype + '_color'],
            'xlabel':opt['xlabel'],
            'ylabel':opt['ylabel'],
        })

    return plotlist


def _get_bounds_plotlist(obj, scale=None):
    if scale is None:scale = 1.0
    plotlist = []

    # Plot the optic extent as taken from the xsize and ysize.
    opt_x = obj.param['xsize'] / 2 * scale
    opt_y = obj.param['ysize'] / 2 * scale
    plotlist.append({
        'axes':'scatter',
        'x':[-1 * opt_x, opt_x, opt_x, -1 * opt_x, -1 * opt_x],
        'y':[opt_y, opt_y, -1 * opt_y, -1 * opt_y, opt_y],
        'linestyle':'--',
        'color':'gray',
    })

    return plotlist


def _get_aperture_plotlist(obj, scale=None):
    if scale is None:scale = 1.0
    plotlist = []

    if (not 'aperture' in obj.param) or (obj.param['aperture'] is None):
        return []

    for apt in obj.param['aperture']:
        apt = xicsrt_aperture._aperture_defaults(apt)
        shape = apt['shape']
        if shape == 'square':
            size = apt['size'][0] / 2 * scale
            origin = apt['origin'].copy() * scale
            x = np.array([-1, 1, 1, -1, -1]) * size + origin[0]
            y = np.array([1, 1, -1, -1, 1]) * size + origin[1]

        if shape == 'rectangle':
            size = apt['size'].copy() / 2 * scale
            origin = apt['origin'].copy() * scale
            x = np.array([-1, 1, 1, -1, -1]) * size[0] + origin[0]
            y = np.array([1, 1, -1, -1, 1]) * size[1] + origin[1]

        if shape == 'triangle':
            vert = apt['vertices'].copy() * scale
            origin = apt['origin'].copy() * scale
            x = np.array([vert[0, 0], vert[1, 0], vert[2, 0], vert[0, 0]]) + origin[0]
            y = np.array([vert[0, 1], vert[1, 1], vert[2, 1], vert[0, 1]]) + origin[0]
        else:
            log.warning(f'Plotting of {shape} aperture not yet implemented.')
            return []

        plotlist.append({
            'axes':'scatter',
            'x':x,
            'y':y,
            'linestyle':'-',
            'color':'black',
        })
    return plotlist


def _update_lim_aspect(ax):
    """
    Update the data limits (xlim & ylim) to produce an equal aspect given a
    fixed plot size.
    """
    xlim = ax.get_xlim()
    xspan = xlim[1] - xlim[0]
    ylim = ax.get_ylim()
    yspan = ylim[1] - ylim[0]
    h = ax.get_window_extent().height
    w = ax.get_window_extent().width
    w_aspect = h / w
    d_aspect = yspan / xspan

    if d_aspect < w_aspect:
        ylim_new = np.array([-0.5, 0.5]) * xspan * w_aspect + sum(ylim) / 2
        ax.set_ylim(ylim_new)
    else:
        xlim_new = np.array([-0.5, 0.5]) * yspan / w_aspect + sum(xlim) / 2
        ax.set_xlim(xlim_new)


def _on_ylims_change(event_ax):
    """
    An Axes callback to update the data limits after a change in the the data
    limits. This is primarily meant to allow retention af a fixed aspect
    after the user has zoomed into a region.
    """

    # We need to disconnect the callback before calling set_ylim
    # Otherwise we will end up in an infinite loop.
    cid_list = list(event_ax.callbacks.callbacks['ylim_changed'].keys())
    for cid in cid_list:
        event_ax.callbacks.disconnect(cid)

    _update_lim_aspect(event_ax)

    event_ax.callbacks.connect('ylim_changed', _on_ylims_change)


def _truncate_mask(mask, max_num):
    if max_num is not None:
        max_num = int(max_num)
        num_mask = np.sum(mask)
        if num_mask > max_num:
            w = np.flatnonzero(mask)
            np.random.shuffle(w)
            mask[w[max_num:]] = False

    return mask


def plot_image(
        results,
        name=None,
        section=None,
        options=None,
        **kwargs,
        ):
    """
    Plot an intersection image along with column and row summation plots.
    """

    if options is None: options = {}
    options.update(kwargs)
    opt = options

    opt.setdefault('name', name)

    config = results['config']
    name = opt['name']

    # Get the crystal object from the dispatcher.
    obj = xicsrt_public.get_element(config, name, section)
    image = results['total']['image'][name]

    opt['pixel_size'] = obj.param['pixel_size']
    opt['size'] = (obj.param['xsize'], obj.param['ysize'])

    fig = detview.view(image, opt)
    controls = detview.add_controls(fig)
    return fig, controls



# -*- coding: utf-8 -*-
"""
.. Authors:
    Novimir Pablant <npablant@pppl.gov>

Define the :class:`ShapePlane` class.
"""

import numpy as np
from copy import deepcopy


from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.optics._ShapeObject import ShapeObject
from xicsrt.tools import xicsrt_math as xm


@dochelper
class ShapePlane(ShapeObject):
    """
    A planar shape.
    This class defines intersections with a plane.
    """

    def intersect(self, rays):
        dist, mask = self.intersect_distance(rays)
        xloc = self.location_from_distance(rays, dist, mask)
        norm = self.intersect_normal(xloc, mask)

        return xloc, norm, mask

    def intersect_distance(self, rays):
        """
        Calculate the distance to an intersection with a plane.
        """

        O = rays['origin']
        D = rays['direction']
        m = rays['mask']

        if self.param['trace_local']:
            distance = np.full(m.shape, np.nan, dtype=np.float64)
            distance[m] = (np.dot((0 - O[m]), np.array([0., 0., 1.]))
                           / np.dot(D[m], np.array([0., 0., 1.])))
        else:
            distance = np.full(m.shape, np.nan, dtype=np.float64)
            distance[m] = (np.dot((self.param['origin'] - O[m]), self.param['zaxis'])
                           / np.dot(D[m], self.param['zaxis']))

        # Update the mask to only include positive distances.
        m &= (distance >= 0)

        return distance, m

    def intersect_normal(self, xloc, mask):
        """
        The planar optic is flat, so the normal direction is always the zaxis.
        """
        m = mask
        norm = np.full(xloc.shape, np.nan, dtype=np.float64)
        norm[m] = self.param['zaxis']
        return norm


# -*- coding: utf-8 -*-
"""
.. Authors:
    Novimir Pablant <npablant@pppl.gov>

Define the :class:`GeometryObject` class.
"""

import numpy as np

# I need to reserve the name 'copy' for keywords.
import copy as m_copy

from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.objects._ConfigObject import ConfigObject
from xicsrt.tools import xicsrt_math as xm

@dochelper
class GeometryObject(ConfigObject):
    """
    The base class for any geometrical objects used in XICSRT.
    """

    def __getattr__(self, key):
        """
        Setup shortcuts for the basic object properties.
        """
        if key == 'xaxis':
            return self.orientation[0, :]
        elif key == 'yaxis':
            return self.orientation[1, :]
        elif key == 'zaxis':
            return self.orientation[2, :]
        else:
            raise AttributeError(key)

    def default_config(self):
        """
        origin
           The x,y,x origin of this element in global coordinates.

        zaxis
           A unit-vector defining the z-axis of the element in global coordinates.
           For most optics: z-axis defines the surface normal direction.

        xaxis (optional)
           A unit-vector defining the x-axis of the element in global coordinates.
           For most optics: x-axis defines the 'width' direction.

           If xaxis is not provided it will be automatically generated by:
           cross(zaxis, [0,1,0]).

           The `yaxis` is always automatically generated and defined by:
           cross(zaxis, xaxis)
        """
        config = super().default_config()
        config['origin'] = np.array([0.0, 0.0, 0.0])
        config['zaxis'] = np.array([0.0, 0.0, 1.0])
        config['xaxis'] = None

        return config

    def check_config(self):

        # I should add a check that zaxis and xaxis are normalized.
        # The other option is just quietly normalize, but that has a risk
        # of confusion to the user.

        # Check here that xaxis and zaxis are orthogonal.
        if self.config['xaxis'] is not None:
            zaxis = np.array(self.config['zaxis'])
            xaxis = np.array(self.config['xaxis'])
            if not np.isclose(np.dot(zaxis, xaxis), 0.0):
                raise ValueError('zaxis and xaxis are not orthogonal.')

    def setup(self):
        super().setup()

        self.param['origin'] = np.array(self.param['origin'])
        self.param['zaxis'] = np.array(self.param['zaxis'])
        if self.param['xaxis'] is None:
            self.param['xaxis'] = self.get_default_xaxis(self.param['zaxis'])
        else:
            self.param['xaxis'] = np.array(self.param['xaxis'])

        # Location with respect to the external coordinate system.
        self.origin = self.param['origin']
        self.set_orientation(self.param['zaxis'], self.param['xaxis'])

    def set_orientation(self, zaxis, xaxis=None):
        if xaxis is None:
            xaxis = self.get_default_xaxis(zaxis)

        self.orientation = np.array([xaxis, np.cross(zaxis, xaxis), zaxis])

    def get_default_xaxis(self, zaxis):
        """
        Get the X-axis using a default definition.

        In order to fully define the orientation of a component both, a z-axis
        and an x-axis are expected.  For certain types of components the x-axis
        definition is unimportant and can be defined using a default definition.
        """

        xaxis = np.cross(np.array([0.0, 0.0, 1.0]), zaxis)
        if not np.all(xaxis == 0.0):
            xaxis /= np.linalg.norm(xaxis)
        else:
            xaxis = np.array([1.0, 0.0, 0.0])

        return xaxis

    def ray_to_external(self, ray_local, copy=False):

        if copy:
            # Programming Note:
            #   If XICSRT is ever updated to always use ray objects then the
            #   copy method should be used instead of deepcopy for speed.
            ray_external = m_copy.deepcopy(ray_local)
        else:
            ray_external = ray_local

        ray_external['origin'] = self.point_to_external(ray_external['origin'])
        ray_external['direction'] = self.vector_to_external(ray_external['direction'])
        return ray_external

    def ray_to_local(self, ray_external, copy=False):
        if copy:
            ray_local = m_copy.deepcopy(ray_external)
        else:
            ray_local = ray_external

        ray_local['origin'] = self.point_to_local(ray_local['origin'])
        ray_local['direction'] = self.vector_to_local(ray_local['direction'])
        return ray_local

    def point_to_external(self, point_local, copy=False):
        return self.vector_to_external(point_local, copy=copy) + self.origin

    def point_to_local(self, point_external, copy=False):
        return self.vector_to_local(point_external - self.origin, copy=copy)

    def vector_to_external(self, vector, copy=False):
        if copy:
            vector = m_copy.copy(vector)

        vector = self.to_ndarray(vector)
        if vector.ndim == 2:
            vector[:] = np.einsum('ij,ki->kj', self.orientation, vector)
        elif vector.ndim == 1:
            vector[:] = np.einsum('ij,i->j', self.orientation, vector)
        else:
            raise Exception('vector.ndim must be 1 or 2')

        return vector

    def vector_to_local(self, vector, copy=False):
        if copy:
            vector = m_copy.copy(vector)

        vector = self.to_ndarray(vector)
        if vector.ndim == 2:
            vector[:] = np.einsum('ji,ki->kj', self.orientation, vector)
        elif vector.ndim == 1:
            vector[:] = np.einsum('ji,i->j', self.orientation, vector)
        else:
            raise Exception('vector.ndim must be 1 or 2')
        return vector

    def aim_to_point(self, aim_point, xaxis=None):
        """
        Set the Z-Axis to aim at a particular point.
        """

        zaxis = aim_point - self.origin
        xm.normalize(zaxis)

        if xaxis is None:
            xaxis = self.get_default_xaxis(zaxis)

        output = {'zaxis': zaxis, 'xaxis': xaxis}

        return output

    def to_ndarray(self, vector_in):
        if not isinstance(vector_in, np.ndarray):
            vector_in = np.array(vector_in, dtype=np.float64)
        return vector_in

    def to_vector_array(self, vector_in):
        """
        Convert a vector to a numpy vector array (if needed).
        """
        vector_in = self.to_ndarray(vector_in)

        if vector_in.ndim < 2:
            return vector_in[None, :]
        else:
            return vector_in




# -*- coding: utf-8 -*-
"""
.. Authors
    Novimir Pablant <npablant@pppl.gov>
    James Kring <jdk0026@tigermail.auburn.edu>
    Yevgeniy Yakusevich <eugenethree@gmail.com>

Contains the XicsrtOpticGeneric class.
"""
from PIL import Image
import numpy as np

from xicsrt.util import profiler
from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.objects._GeometryObject import GeometryObject

from xicsrt.tools import xicsrt_aperture

@dochelper
class TraceObject(GeometryObject):
    """
    A generic optical element and base class for raytracing objects in XICSRT.

    Optical elements have a position and rotation in 3D space and a finite 
    extent. Additional properties, such as as crystal spacing, rocking curve, 
    and reflectivity, should be defined in derived classes.
    """

    def default_config(self):
        """
        xsize
          The size of this element along the xaxis direction.
          Typically corresponds to the 'width' of the optic.

        ysize
          The size of this element along the yaxis direction.
          Typically corresponds to the 'height' of the optic.

        zsize
          The size of this element along the zaxis direction.
          Typically not required, but if needed will correspond to the 'depth'
          of the optic.

        pixel_size: float (None)
          The pixel size, used for binning rays into images.
          This is currently a single number signifying square pixels.

        aperture: dict or array (None)
          Define one or more apertures to to apply to this optic.
          Each aperture is defined as a dictionary with the following keys:
          shape, size, origin, logic. The origin and logic field keys are
          optional. The interpretation of size will depend on the provided
          shape.

        trace_local: bool (False)
          If true: transform rays to optic local coordinates before raytracing,
          do raytracing in local coordinates, then transform back to global
          coordinates.

          The default is 'false' as most built-in optics can perform raytracing
          in global coordinates. This option is convenient for optics with
          complex geometry for which intersection and reflection equations
          are easier or more clear to program in fixed local coordinates.

        check_size: bool (true)
          Perform a check for whether the rays intersect the optic within the
          defined bounds (usually defined by 'xsize' and 'ysize'). If set to
          `False` all rays with a defined reflection/transmission condition
          will be traced if an intersection can be determined.

        check_aperture: bool (true)
          Perform a check for whether the rays intersect the optic within the
          defined bounds (usually defined by 'xsize' an 'ysize'). If set to
          `False` all rays with a defined reflection/transmission condition
          will be traced if an intersection can be determined.

        filters
          No documentation yet. Please help improve XICSRT!
        """
        config = super().default_config()
        
        # spatial information
        config['xsize']          = None
        config['ysize']          = None
        config['zsize']          = None
        config['pixel_size']     = None

        # boolean settings
        config['trace_local'] = False
        config['check_size'] = True
        config['check_aperture'] = True
        
        # Aperture list
        config['aperture'] = None

        # Filters
        config['filters'] = []

        return config

    def initialize(self):
        super().initialize()

        if self.param['xsize'] and self.param['ysize']:
            # autofill pixel grid sizes
            if self.param['pixel_size'] is None:
                self.param['pixel_size'] = self.param['xsize']/100

            # Determine the number of pixels on the detector.
            # For now assume that the user set the width of the detector to be
            # a multiple of the pixel size.
            #
            # Except for the detector there is really no reason that this would
            # always be true, so for now only make this a warning. I need to think
            # about how to handle this better.
            pixel_xsize = self.param['xsize'] / self.param['pixel_size']
            pixel_ysize = self.param['ysize'] / self.param['pixel_size']
            try:
                np.testing.assert_almost_equal(pixel_xsize, np.round(pixel_xsize))
                np.testing.assert_almost_equal(pixel_ysize, np.round(pixel_ysize))
            except AssertionError:
                self.log.warning(f"Optic width ({self.param['xsize']:0.4f}x{self.param['ysize']:0.4f})"
                                 f"is not a multiple of the pixel_size ({self.param['pixel_size']:0.4f})."
                                 f"May lead to truncation of output image."
                                 )

            self.param['pixel_xsize'] = int(np.round(pixel_xsize))
            self.param['pixel_ysize'] = int(np.round(pixel_ysize))
            self.log.debug(f"Pixel grid size: {self.param['pixel_xsize']} x {self.param['pixel_ysize']}")

            self.param['enable_image'] = True
        else:
            self.param['enable_image'] = False

    def trace_global(self, rays):
        """
        This is method that is called by the dispacher to perform ray-tracing
        for this optic. Rays into and out of this method are always in global
        coordinates.

        It may be convenient for some optics object to do raytracing in local
        coordinates rather than in global coordinates. This method facilitates
        this by implementing the 'trace_local' configuration option.
        """

        if self.param['trace_local']:
            self.log.debug('Converting to local coordinates.')
            rays = self.ray_to_local(rays, copy=False)

        rays = self.trace(rays)

        if self.param['trace_local']:
            self.log.debug('Converting to external coordinates.')
            rays = self.ray_to_external(rays, copy=False)
        return rays

    def trace(self, rays):
        """
        The main method that performs raytracing for this optic.

        Raytracing here may be done in global or local coordinates depending on
        the how the optic is designed and the value of the configuration option:
        'trace_local'.
        """

        xloc, norm, mask = self.intersect(rays)
        mask = self.check_bounds(xloc, mask)
        self.log.debug('Rays on {}: {:0.4e}'.format(self.__class__.__name__, np.sum(mask)))
        rays = self.interact(rays, xloc, norm, mask)
        self.log.debug('Rays from {}: {:0.4e}'.format(self.__class__.__name__, np.sum(mask)))

        return rays

    def intersect(self, rays):
        raise NotImplementedError('This method should be reimplemented in a ShapeObject.')

    def interact(self, rays, xloc, norm, mask):
        raise NotImplementedError('This method should be reimplemented in a IntersectObject.')

    def check_bounds(self, X, mask):
        m = mask

        if self.param['trace_local']:
            X_local = X
        else:
            X_local = np.zeros(X.shape, dtype=np.float64)
            X_local[m] = self.point_to_local(X[m])

        mask = self.check_size(X_local, mask)
        mask = self.check_aperture(X_local, mask)

        return mask

    def check_size(self, X_local, mask):
        """
        Check if the ray intersection is within the optic bounds as set
        by the xsize, ysize and zsize config options.

        Note:
            This method expects to be given the ray intersections in local
            coordinates. Generally this method should not be called directly,
            instead use `check_bounds`.
        """
        m = mask

        if self.param['check_size']:
            if self.param['xsize'] is not None:
                m[m] &= (np.abs(X_local[m,0]) < self.param['xsize'] / 2)
            if self.param['ysize'] is not None:
                m[m] &= (np.abs(X_local[m,1]) < self.param['ysize'] / 2)
            if self.param['zsize'] is not None:
                m[m] &= (np.abs(X_local[m,2]) < self.param['zsize'] / 2)
        
        return mask
    
    def check_aperture(self, X_local, mask):
        """
        Check if the ray intersection is within the aperture as set
        by the 'aperture' config option.

        Note:
            This method expects to be given the ray intersections in local
            coordinates. Generally this method should not be called directly
            instead use `check_bounds`.
        """
        m = mask

        if self.param['check_aperture']:
            m_aperture = xicsrt_aperture.aperture_mask(X_local, m, self.param['aperture'])
            m[m] = m_aperture[m]

        return mask

    def make_image(self, rays):
        """
        Collect the rays that intersect with this optic into a pixel array that
        can be used to generate an intersection image.

        Programming Notes
        -----------------

        It is important that this calculation is compatible with intersect_check
        in terms of floating point errors.  The simple way to achieve this is
        to ensure that both use the same calculation method.
        """

        if not self.param['enable_image']:
            return None

        image = np.zeros((self.param['pixel_xsize'], self.param['pixel_ysize']))
        X = rays['origin']
        m = rays['mask'].copy()
        
        num_lines = np.sum(m)
        
        # Add the ray hits to the pixel array
        if num_lines > 0:
            # Transform the intersection coordinates from external coordinates
            # to local optical 'pixel' coordinates.
            point_loc = self.point_to_local(X[m])
            pix = point_loc / self.param['pixel_size']

            # Convert from pixels to channels.
            # The channel coordinate is defined from the *center* of the bottom left
            # pixel. The pixel coordinate is define from the geometrical center of
            # the detector (this could be in the middle of or in between pixels).
            channel = np.zeros(pix.shape)
            channel[:,0] = pix[:,0] + (self.param['pixel_xsize'] - 1)/2
            channel[:,1] = pix[:,1] + (self.param['pixel_ysize'] - 1)/2
            
            # Bin the channels into integer values so that we can use them as
            # indexes into the image. Keep in mind that channel coordinate
            # system is defined from the center of the pixel.
            channel = np.round(channel).astype(int)
            
            # Check for any hits that are outside of the image.
            # These are possible due to floating point calculations.
            m = np.ones(num_lines, dtype=bool)
            m &= channel[:,0] >= 0
            m &= channel[:,0] < self.param['pixel_xsize']
            m &= channel[:,1] >= 0
            m &= channel[:,1] < self.param['pixel_ysize']
            num_out = np.sum(~m)
            if num_out > 0:
                self.log.warning('Rays found outside of pixel grid ({}).'.format(num_out))
            
            # I feel like there must be a faster way to do this than to loop over
            # every intersection.  This could be slow for large arrays.
            for ii in range(num_lines):
                if m[ii]:
                    image[channel[ii,0], channel[ii,1]] += 1

        return image



# -*- coding: utf-8 -*-
"""
Authors
-------
- Novimir A. Pablant <npblant@pppl.gov>
"""

import numpy as np
import logging
import copy

class RayArray(dict):
    """
    The base class for an Ray array.

    The RayArray object is essentially a dictionary of numpy arrays.
    Some convenience methods have been added.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if 'origin' in self and 'direction' in self:
            self.initialize()

    def initialize(self):
        """
        Initialize the ray array.
        This will ensure that all properties are present and of the correct type.
        """

        if not isinstance(self['origin'], np.ndarray):
            self['origin'] = np.array(self['origin'])
        if not isinstance(self['direction'], np.ndarray):
            self['direction'] = np.array(self['direction'])

        if not (('origin' in self) and ('direction' in self)):
            raise Exception('Cannot initialize, origin and direction must be present.')

        shape = self['origin'].shape
        if not 'mask' in self:
            self['mask'] = np.ones(shape[0], dtype=bool)
        if not 'wavelength' in self:
            self['wavelength'] = np.zeros(shape[0])

        if not isinstance(self['mask'], np.ndarray):
            self['mask'] = np.array(self['mask'])
        if not isinstance(self['wavelength'], np.ndarray):
            self['wavelength'] = np.array(self['wavelength'])

    def __getattribute__(self, key):
        """
        Setup shortcuts for the basic ray properties.
        """
        if key == 'O' or key == 'origin':
            return self['origin']
        elif key == 'D' or key == 'direction':
            return self['direction']
        elif key == 'W' or key == 'wavelength':
            return self['wavelength']
        elif key == 'M' or key == 'mask':
            return self['mask']
        else:
            return super().__getattribute__(key)

    def __setattr__(self, key, value):
        """
        Setup shortcuts for the basic ray properties.
        """

        if key == 'O' or key == 'origin':
            self['origin'] = value
        elif key == 'D' or key == 'direction':
            self['direction'] = value
        elif key == 'W' or key == 'wavelength':
            self['wavelength'] = value
        elif key == 'M' or key == 'mask':
            self['mask'] = value
        else:
            super().__setattr(key, value)

    def zeros(self, num):
        self['origin'] = np.zeros((num, 3))
        self['direction'] = np.zeros((num, 3))
        self['mask'] = np.zeros((num), dtype=bool)
        self['wavelength'] = np.zeros((num))

    def copy(self):
        ray_new = RayArray()
        for key in self:
            ray_new[key] = self[key].copy()
        return ray_new

    def extend(self, ray_in):
        for key in self:
            self[key] = np.concatenate((self[key], ray_in[key]))



# -*- coding: utf-8 -*-
"""
.. Authors
    Novimir Pablant <npablant@pppl.gov>
    James Kring <jdk0026@tigermail.auburn.edu>
    Yevgeniy Yakusevich <eugenethree@gmail.com>

A set of mathematical utilities and vector convenience functions for XICSRT.
"""

import numpy as np


def distance_point_to_line(origin, normal, point):
    """
    Find the closest distnace between a point and a line in 3D.
    """
    o = origin
    n = normal
    p = point
    t = np.dot(p - o, n) / np.dot(n, n)
    d = np.linalg.norm(np.outer(t, n) + o - p, axis=1)
    return d


def intersect_ray_plane(ray, plane):
    """
    Find the intersection between a ray and a plane in 3D.
    """

    if ray['origin'].ndim == 1:
        distance = (np.dot((plane['origin'] - ray['origin']), plane['zaxis'])
                       / np.dot(ray['direction'], plane['zaxis']))
        intersect = ray['origin'] + ray['direction'] * distance
    else:
        raise NotImplementedError('Only a single ray is currently supported.')

    return intersect


def toarray_1d(a):
    """
    Convert the input to a ndarray with at least 1 dimension.
    This is similar to the numpy function atleast_1d, but has less overhead
    and is jax compatible.
    """
    a = np.asarray(a)
    if a.ndim == 0:
        a = a.reshape(1)
    return a


def vector_angle(a, b):
    """
    Find the angle between two vectors.
    """

    if a.ndim == 2 & b.ndim == 2:
        a_mod = np.linalg.norm(a, axis=1)
        b_mod = np.linalg.norm(b, axis=1)
        dot = np.einsum('ij,i,ij,i->i', a, 1/a_mod, b, 1/b_mod, optimize=True)
    elif a.ndim == 1 & b.ndim == 1:
        a_mod = np.linalg.norm(a)
        b_mod = np.linalg.norm(b)
        dot = np.dot(a/a_mod, b/b_mod)
    else:
        raise Exception('Input must have 1 or 2 dimensions.')
    angle = np.arccos(dot)
    return angle


def vector_rotate(a, b, theta):
    """
    Rotate vector a around vector b by an angle theta (radians)

    Programming Notes:
      u: parallel projection of a on b_hat.
      v: perpendicular projection of a on b_hat.
      w: a vector perpendicular to both a and b.
    """
    a = np.asarray(a)
    b = np.asarray(b)

    if a.ndim == 2:
        b_hat = b / np.linalg.norm(b)
        dot = np.einsum('ij,j->i', a, b_hat, optimize=True)
        u = np.einsum('i,j->ij', dot, b_hat, optimize=True)
        v = a - u
        w = np.cross(b_hat, v)
        c = u + v * np.cos(theta) + w * np.sin(theta)
    elif a.ndim == 1:
        b_hat = b / np.linalg.norm(b)
        u = b_hat * np.dot(a, b_hat)
        v = a - u
        w = np.cross(b_hat, v)
        c = u + v * np.cos(theta) + w * np.sin(theta)
    else:
        raise Exception('Input array must be 1d (vector) or 2d (array of vectors)')
    return c


def magnitude(vector):
    """
    Calculate magnitude of a vector or array of vectors.
    """
    if vector.ndim > 1:
        mag = np.linalg.norm(vector, axis=1)
    else:
        mag = np.linalg.norm(vector)

    return mag


def normalize(vector):
    """
    Normalize a vector or an array of vectors.
    If an array of vectors is given it should have the shape (N,M) where
    |  N: Number of vectors
    |  M: Vector length
    """

    if vector.ndim > 1:
        norm = np.linalg.norm(vector, axis=1)
        vector /= np.expand_dims(norm, 1)
    else:
        norm = np.linalg.norm(vector)
        vector /= norm

    return vector


def sinusoidal_spiral(phi, b, r0, theta0):
    r = r0 * (np.sin(theta0 + (b-1)*phi)/np.sin(theta0))**(1/(b-1))
    return r


def rotation_matrix(axis, theta):
    """
    Return the rotation matrix associated with counterclockwise rotation about
    the given axis by theta radians.
    """
    axis = np.asarray(axis)
    axis = axis/np.sqrt(np.dot(axis, axis))
    a = np.cos(theta/2.0)
    b, c, d = -axis*np.sin(theta/2.0)
    aa, bb, cc, dd = a*a, b*b, c*c, d*d
    bc, ad, ac, ab, bd, cd = b*c, a*d, a*c, a*b, b*d, c*d
    matrix = np.array(
        [[aa + bb - cc - dd, 2 * (bc + ad), 2 * (bd - ac)]
        ,[2 * (bc - ad), aa + cc - bb - dd, 2 * (cd + ab)]
        ,[2 * (bd + ac), 2 * (cd - ab), aa + dd - bb - cc]])

    return matrix


def bragg_angle(wavelength, crystal_spacing):
    """
    The Bragg angle calculation is used so often that it deserves its own
    function.

    .. Note::
      The crystal_spacing here is the true spacing, not the 2d spacing that
      is sometimes used in the literature.
    """
    bragg_angle = np.arcsin(wavelength / (2 * crystal_spacing))
    return bragg_angle


def cyl_from_car(point_car):
    """
    Convert from cartesian to cylindrical coordinates.
    """
    if not isinstance(point_car, np.ndarray):
        point_car = np.array(point_car)

    point_cyl = np.empty(point_car.shape)

    if point_car.ndim == 2:
        point_cyl[:, 0] = np.sqrt(np.sum(point_car[:, 0:2]**2, 1))
        point_cyl[:, 1] = np.arctan2(point_car[:, 1], point_car[:, 0])
        point_cyl[:, 2] = point_car[:, 2]
    else:
        point_cyl[0] = np.sqrt(np.sum(point_car[0:2]**2))
        point_cyl[1] = np.arctan2(point_car[1], point_car[0])
        point_cyl[2] = point_car[2]

    return point_cyl


def car_from_cyl(point_cyl):
    """
    Convert from cylindrical to cartesian coordinates.
    """
    if not isinstance(point_cyl, np.ndarray):
        point_cyl = np.array(point_cyl)

    point_car = np.empty(point_cyl.shape)

    if point_cyl.ndim == 2:
        point_car[:, 0] = point_cyl[:, 0]*np.cos(point_cyl[:, 1])
        point_car[:, 1] = point_cyl[:, 0]*np.sin(point_cyl[:, 1])
        point_car[:, 2] = point_cyl[:, 2]
    else:
        point_car[0] = point_cyl[0]*np.cos(point_cyl[1])
        point_car[1] = point_cyl[0]*np.sin(point_cyl[1])
        point_car[2] = point_cyl[2]

    return point_car


def tor_from_car(point_car, major_radius):
    """
    Convert from cartesian to toroidal coordinates.

    Arguments
    ---------
    point_car: array [meters]
      Cartesian coordinates [x,y,z]

    major_radius: float [meters]
      Torus Major Radius

    Returns
    -------
    point_tor: array [meters]
      Toroidal coordinates [r_min, theta_poloidal, theta_toroidal]
    """
    if not isinstance(point_car, np.ndarray):
        point_car = np.array(point_car)

    point_tor = np.empty(point_car.shape)

    if point_tor.ndim == 2:
        d = np.linalg.norm(point_car[:, 0:2], axis=1) - major_radius
        point_tor[:, 2] = np.arctan2(point_car[:, 1], point_car[:, 0])
        point_tor[:, 1] = np.arctan2(point_car[:, 2], d)
        point_tor[:, 0] = np.sqrt(np.power(point_car[:, 2], 2) + np.power(d, 2))
    else:
        d = np.linalg.norm(point_car[0:2]) - major_radius
        point_tor[2] = np.arctan2(point_car[1], point_car[0])
        point_tor[1] = np.arctan2(point_car[2], d)
        point_tor[0] = np.sqrt(np.power(point_car[2], 2) + np.power(d, 2))

    return point_tor


def car_from_tor(point_tor, major_radius):
    """
    Convert from toroidal to cartesian coordinates.

    Arguments
    ---------
    point_tor: array [meters]
      Toroidal coordinates [r_min, theta_poloidal, theta_toroidal]

    major_radius: float [meters]
      Torus Major Radius

    Returns
    -------
    point_car: array [meters]
      Cartesian coordinates [x,y,z]
    """

    if not isinstance(point_tor, np.ndarray):
        point_tor = np.array(point_tor)

    point_car = np.empty(point_tor.shape)

    if point_tor.ndim == 2:
        point_car[:, 0] = major_radius*np.cos(point_tor[:, 2])
        point_car[:, 1] = major_radius*np.sin(point_tor[:, 2])
        vector = point_car[:, 0:2]/np.linalg.norm(point_car[:, 0:2], axis=1)

        point_car[:, 0] = point_car[:, 0] + vector[:, 0]*point_tor[:, 0]*np.cos(point_tor[:, 1])
        point_car[:, 1] = point_car[:, 1] + vector[:, 1]*point_tor[:, 0]*np.cos(point_tor[:, 1])
        point_car[:, 2] = point_tor[:, 0]*np.sin(point_tor[:, 1])
    else:
        point_car[0] = major_radius*np.cos(point_tor[2])
        point_car[1] = major_radius*np.sin(point_tor[2])
        vector = point_car[0:2]/np.linalg.norm(point_car[0:2])

        point_car[0] = point_car[0] + vector[0]*point_tor[0]*np.cos(point_tor[1])
        point_car[1] = point_car[1] + vector[1]*point_tor[0]*np.cos(point_tor[1])
        point_car[2] = point_tor[0]*np.sin(point_tor[1])

    return point_car


def point_in_triangle_2d(pt, p0, p1, p2):
    """
    Determine if a point (or set of points) fall within a triangle as specified
    by three vertices.  Calculation is performed in two dimensions (2D).
    """
    pt = np.asarray(pt)
    if pt.ndim == 2:
        area = 0.5 * (-p1[1] * p2[0] + p0[1] * (-p1[0] + p2[0]) + p0[0] * (p1[1] - p2[1]) + p1[0] * p2[1])
        a = 1 / (2 * area) * (p0[1] * p2[0] - p0[0] * p2[1] + (p2[1] - p0[1]) * pt[:, 0] + (p0[0] - p2[0]) * pt[:, 1])
        b = 1 / (2 * area) * (p0[0] * p1[1] - p0[1] * p1[0] + (p0[1] - p1[1]) * pt[:, 0] + (p1[0] - p0[0]) * pt[:, 1])
        c = 1 - a - b
    else:
        area = 0.5 * (-p1[1] * p2[0] + p0[1] * (-p1[0] + p2[0]) + p0[0] * (p1[1] - p2[1]) + p1[0] * p2[1])
        a = 1 / (2 * area) * (p0[1] * p2[0] - p0[0] * p2[1] + (p2[1] - p0[1]) * pt[0] + (p0[0] - p2[0]) * pt[1])
        b = 1 / (2 * area) * (p0[0] * p1[1] - p0[1] * p1[0] + (p0[1] - p1[1]) * pt[0] + (p1[0] - p0[0]) * pt[1])
        c = 1 - a - b

    return (a >= 0) & (b >= 0) & (c >= 0)


# -*- coding: utf-8 -*-
"""
.. Authors:
    Novimir pablant <npablant@pppl.gov>

An interface to matplotlib that allows specification of complex plots
though a list of parameter dictionaries.

Example
-------

The simplest example:

.. code::

  import numpy as np
  import mirplot

  x = np.arange(10)
  y = x
  plotlist = [{'x':x, 'y':y}]
  fig = mirplot.plot_to_screen(plotlist)

Any supported plot properties can be added to the plot dictionary:

.. code:

    plotlist = [{
        'x':x,
        'y':y,
        'xbound':[0,1],
        'ybound':[0,1],
        'xtitle':'This is the x-axis',
        'ytitle':'This is the y-axis',
        }]
    fig = mirplot.plot_to_screen(plotlist)

To add multiple plots to a single figure add parameter dicts ta the plotlist:

.. code:

    plotlist = [
        {'x':x1, 'y':y1},
        {'x':x2, 'y':y2},
        ]
    fig = mirplot.plot_to_screen(plotlist)

If axes names are provided then plots will be added to separate subfigures
(stacked vertically). Each unique axes name will result in a new subfigure.

.. code:

    plotlist = [
        {'axes':'plot 1', 'x':x1, 'y':y1},
        {'axes':'plot 2', 'x':x2, 'y':y2},
        ]
    fig = mirplot.plot_to_screen(plotlist)

mirplot can also be used with predifined axes. For this purpose the axes must
be placed into a dictionary and passed to `plot_to_axes`.

.. code:

    fig, axs = plt.subplots(1, 2)

    axesdict = {
        'plot 1':axs[0],
        'plot 2':axs[1],
        }
    plotlist = [
        {'axes':'plot 1', 'x':x1, 'y':y1},
        {'axes':'plot 2', 'x':x2, 'y':y2},
        ]

    fig = mirplot2.plot_to_axes(plotlist, axesdict)

mirplot properties
------------------

A set of unique plot and axes properties are defined by mirplot to enable
a complete dictionary definition.

type : str ('line')
  Allowed Values: line, errorbar, scatter, fill_between, hline, vline, hspan,
  vspan.

legend : bool (false)
  Set to true to show the legend in this subplot.

matplotlib properties
---------------------

Any matplotlib plot or axes property that can be set using a simple
`set_prop(value)` method is supported. Certain properties requiring
a more complex set call are also supported.

"""

import logging
import copy

import matplotlib
import numpy as np

m_log = logging.getLogger(__name__.split('.')[-1])
m_log.setLevel(logging.INFO)

__version__ = '2.0.0'
__version_date__ = '2022-05-20'

def plot_to_screen(plotlist, show=True):
    matplotlib.pyplot.ioff()

    namelist = _autoname_plots(plotlist)
    fig = _make_figure(namelist)
    axesdict = _make_axes(namelist, fig)
    plot_to_axes(plotlist, axesdict)

    matplotlib.pyplot.ion()
    if show:
        fig.show()

    return fig


def plot_to_file(plotlist, filename):

    fig = plot_to_screen(plotlist, show=False)
    fig.savefig(filename)
    m_log.info('Saved figure to file: {}'.format(filename))


def plot_to_axes(plotlist, axesdict):

    # The order of these calls is important so we need to use multiple loops

    axeslist = []
    for plot in plotlist:
        if plot.get('type') == 'figure':
            axes = list(axesdict.values())[0]
        else:
            axes = plot['axes']

        if isinstance(axes, str):
            if axes in axesdict:
                axes = axesdict[axes]
            else:
                raise Exception(f'Named axes {axes} not found.')
        axeslist.append(axes)

    for plot in plotlist:
        _set_plot_defaults(plot)
        _clean_plot_prop(plot)

    for ii, plot in enumerate(plotlist):
        _apply_plot_prop(plot, axeslist[ii])

    for ii, plot in enumerate(plotlist):
        _apply_axes_prop(plot, axeslist[ii])

    for ii, plot in enumerate(plotlist):
        _apply_fig_prop(plot, axeslist[ii])


def _set_plot_defaults(prop):
    prop.setdefault('type', 'line')

    if prop.get('type') == 'figure':
        return
    if prop.get('type') == 'axes':
        return

    prop.setdefault('x')
    prop.setdefault('y')
    prop.setdefault('xerr')
    prop.setdefault('yerr')
    prop.setdefault('s', 15)
    prop.setdefault('legend_fontsize', 12.0)
    prop.setdefault('legend_framealpha', 0.7)

    if prop['type'] == 'image':
        if 'extent' not in prop:
            if (prop['x'] is not None) and (prop['y'] is not None):
                pass
                # This needs work. If I assume that x and y define the pixel center
                # then I could automatically calculate an extent here.
                # prop['extent'] = [min(prop['x']), max(prop['x']), min(prop['y']), max(prop['y'])]
    elif prop['type'] == 'hline':
        if prop['y'] is None:
            prop['y'] = [0]
    elif prop['type'] == 'vline':
        if prop['x'] is None:
            prop['x'] = [0]
    else:
        if prop['x'] is None:
            prop['x'] = np.arange(len(prop['y']))
        #if not 'ybound' in prop:
        #    yrange = np.array([np.nanmin(prop['y']), np.nanmax(prop['y'])])
        #    prop['ybound'] = yrange + np.array([-0.1, 0.1]) * (yrange[1] - yrange[0])

def _clean_plot_prop(prop):
    """
    Check the plot properties and cleanup or provides errors.
    """
    if prop.get('type') == 'figure':
        return
    if prop.get('type') == 'axes':
        return

    if 'x' in prop and prop['x'] is not None:
        if np.isscalar(prop['x']):
            prop['x'] = np.asarray([prop['x']])
        else:
            prop['x'] = np.asarray(prop['x'])

    if 'y' in prop and prop['y'] is not None:
        if np.isscalar(prop['y']):
            prop['y'] = np.asarray([prop['y']])
        else:
            prop['y'] = np.asarray(prop['y'])


def _apply_plot_prop(prop, axes):
    if prop.get('type') == 'figure':
        return
    if prop.get('type') == 'axes':
        return

    if prop['type'] == 'line':
        plotobj, = axes.plot(prop['x'], prop['y'])
    elif prop['type'] == 'errorbar':
        plotobj = axes.errorbar(
            prop['x']
            , prop['y']
            , xerr=prop['xerr']
            , yerr=prop['yerr']
            , fmt='none'
            , capsize=prop['capsize'])
    elif prop['type'] == 'scatter':
        plotobj = axes.scatter(prop['x'], prop['y'], s=prop['s'], marker=prop.get('marker', None))
    elif prop['type'] == 'fill_between' or prop['type'] == 'fillbetween':
        plotobj = axes.fill_between(prop['x'], prop['y'], prop['y1'])
    elif prop['type'] == 'hline':
        plotobj = axes.axhline(prop['y'][0])
    elif prop['type'] == 'vline':
        plotobj = axes.axvline(prop['x'][0])
    elif prop['type'] == 'hspan':
        plotobj = axes.axhspan(prop['y'][0], prop['y'][1])
    elif prop['type'] == 'vspan':
        plotobj = axes.axvspan(prop['x'][0], prop['x'][1])
    elif prop['type'] == 'image':
        # Some value (any value) must be given for the extent.
        plotobj = axes.imshow(prop['z'], aspect='auto', extent=prop['extent'])
    else:
        raise Exception('Plot type unknown: {}'.format(prop['type']))

    # Certain plot types actually consist of collections of line objects.
    if prop['type'] == 'errorbar':
        obj_list = []
        obj_list.extend(plotobj[1])
        obj_list.extend(plotobj[2])
    else:
        obj_list = [plotobj]

    # Loop through all objects and set the appropriate properties.
    for key in prop:
        for obj in obj_list:
            obj_dir = dir(obj)
            if key == 'markersize':
                if prop['type'] == 'scatter':
                    sizes = obj.get_sizes()
                    sizes[:] = prop['markersize']
                    obj.set_sizes(sizes)
                else:
                    obj.set_markersize(prop['markersize'])
            else:
                # This will catch any properties can can be simply set with a
                # plot.set_prop(value) function.
                funcname = 'set_' + key
                if funcname in obj_dir:
                    if prop[key] is not None:
                        getattr(obj, funcname)(prop[key])


def _apply_axes_prop(prop, axes):
    if prop.get('type') == 'figure':
        return

    prop = copy.copy(prop)

    # In some cases the order of these statements is important.
    # (For example xscale needs to come before xbound I think.)

    axes.tick_params('both'
                     , direction='in'
                     , which='both'
                     , top='on'
                     , bottom='on'
                     , left='on'
                     , right='on')

    ax_dir = dir(axes)
    for key in prop:
        if key == 'xscale':
            if prop['xscale'] == 'log':
                nonpositive = 'clip'
            else:
                nonpositive = None
            axes.set_xscale(prop['xscale'], nonpositive=nonpositive)
        elif key == 'yscale':
            if prop['yscale'] == 'log':
                nonpositive = 'clip'
            else:
                nonposisitve = None
            axes.set_yscale(prop['yscale'], nonpositive=nonpositive)
        elif key == 'legend':
            axes.legend(loc=prop.get('legend_location')
                        , fontsize=prop.get('legend_fontsize')
                        , framealpha=prop.get('legend_framealpha')
                        )
        elif key == 'label_outer':
            axes.label_outer()
        else:
            # This will catch any properties can can be simply set with a
            # axes.set_prop(value) function.
            funcname = 'set_'+key
            if funcname in ax_dir:
                if prop[key] is not None:
                    getattr(axes, funcname)(prop[key])


def _apply_fig_prop(prop, ax):
    if not prop.get('type') == 'figure':
        return

    fig = ax.figure
    fig_dir = dir(fig)
    for key in prop:
        if key == 'suptitle':
            x = prop.get('suptitle_x', 0.02)
            y = prop.get('suptitle_y', 0.98)
            ha = prop.get('suptitle_ha', 'left')
            weight = prop.get('suptitle_weight')
            fig.suptitle(prop['suptitle'], x=x, y=y, ha=ha, weight=weight)
        else:
            # This will catch any properties can can be simply set with a
            # axes.set_prop(value) function.
            funcname = 'set_' + key
            if funcname in fig_dir:
                if prop[key] is not None:
                    getattr(fig, funcname)(prop[key])


def _make_figure(namelist):
    size = _get_figure_size(len(namelist))
    fig = matplotlib.pyplot.figure(figsize=size)
    return fig


def _make_axes(namelist, fig):

    numaxes = len(namelist)
    axesdict = {}
    for ii, name in enumerate(namelist):
        axesdict[name] = fig.add_subplot(numaxes, 1, ii + 1)

    fig.axesdict = axesdict

    return axesdict


def _autoname_plots(plotlist, sequential=False):
    """
    Automatically name any plots that were not given a name by the user.
    """
    namelist = []

    count = 0
    for plot in plotlist:
        if plot.get('type') == 'figure':
            continue
        name = plot.get('axes')
        if name is None:
            if sequential:
                num = count
            else:
                num = 0
            name = '_autoname_{:02d}'.format(num)
            count += 1
        plot['axes'] = name
        namelist.append(name)

    # Extract the unique names in order.
    namelist = list(dict.fromkeys(namelist))
    return namelist


def _get_figure_size(numaxes):
    """
    Return the default figure size.
    Width: 8 units
    Height: 3 units for every subplot or max 9 units
    Return
    ------

    (width, height)
      The figure size in inches.
    """

    figure_width = 8
    figure_height = max(6, min(numaxes * 3, 10))

    return (figure_width, figure_height)



# -*- coding: utf-8 -*-
"""
.. Authors
    Novimir Pablant <npablant@pppl.gov>

A set of routines for related to Voigt distributions.
"""

import numpy as np
from scipy.interpolate import interp1d
from scipy.special import wofz

def voigt(
        x
        ,intensity=None
        ,location=None
        ,sigma=None
        ,gamma=None):
    """
    The Voigt function is also the real part of  w(z) = exp(-z^2) erfc(iz), 
    the complex probability function, which is also known as the Faddeeva 
    function. Scipy has implemented this function under the name wofz()
    """
    
    z = (x - location + 1j*gamma)/np.sqrt(2)/sigma
    y = wofz(z).real/np.sqrt(2*np.pi)/sigma * intensity
    return y


def voigt_cdf_tab(gamma, sigma, gridsize=None, cutoff=None):

    # This is a numerical method to calculate the cumulative distribution fuction
    # for a voigt profile.  This works reasonably well, but is limited both by
    # the sampling resolution, and the chosen bounds.
    # 
    # In this case the CDF is calculated with a variable grid density to help
    # mitigate those effects.
    #
    # There are a couple of possibilities to speed this up:
    #  1. It may be possible to optimize the grid spacing by using
    #     a fuction that increases faster away from zero.
    #  2. The CDF is symetric so only calculation up to x=0 is needed.
    #  3. For some applicaiton I might be able to use a psudo-voigt
    #     calculation that may be faster than the wofz implementation
    #     (at the expense of accuracy.)
    

    if gridsize is None: gridsize = 1000
    if cutoff is None: cutoff = 1e-5
        
    # The current scheme works well with a minimum of 100 points.
    # It is possible to go as low as 50 points, but accuracy is not great.
    gridsize_min = 100

    fraction = 0.5
    gauss_hwfm = np.sqrt(2.0*np.log(1.0/fraction))*sigma
    lorentz_hwfm = gamma*np.sqrt(1.0/fraction - 1.0)
    # This is always larger than the voigt hwfm (half width at percentile).
    hwfm_max = np.sqrt(gauss_hwfm**2 + lorentz_hwfm**2)

    min_spacing = hwfm_max/5.0
    value = gridsize_min/2*min_spacing

    # Determine a cutoff value.
    lorentz_cutoff = gamma*np.sqrt(1.0/cutoff - 1.0)
    gauss_cutoff = np.sqrt(-1 * sigma**2 * 2 * np.log(cutoff*sigma*np.sqrt(2*np.pi)))
    value_cutoff = max(lorentz_cutoff, gauss_cutoff)
    base = np.exp(1/10 * np.log(value_cutoff/value))

    bounds = np.linspace(-value, value, gridsize+1)
    bounds = bounds*base**np.abs(bounds/value*10)
    cdf_x = (bounds[:-1]+bounds[1:])/2
        
    # We must use a properly normalized voigt here (intensity=1.0)
    cdf_y = voigt(
        cdf_x
        ,intensity=1.0
        ,location=0.0
        ,sigma=sigma
        ,gamma=gamma)

    cdf_ydx = (cdf_y*(bounds[1:]-bounds[:-1]))
    cdf = np.cumsum(cdf_ydx)

    # These checks are only useful if the user changes the number
    # of calculated points.
    if (np.sum((cdf > 0.25) & (cdf < 0.75)) < 3):
        raise Exception('Voight CDF calculation does not have enough resolution.')
    if (np.max(cdf) < 0.99):
        raise Exception('Voight CDF calculation domain too small.')
    
    return bounds[1:], cdf


def voigt_cdf_interp(gamma, sigma, gridsize=None):
    x, cdf = voigt_cdf_tab(gamma, sigma, gridsize)
    interp = interp1d(x, cdf, kind='quadratic')
    return interp


def voigt_invcdf_interp(gamma, sigma, gridsize=None):
    x, cdf = voigt_cdf_tab(gamma, sigma, gridsize)
    interp = interp1d(cdf, x, kind='quadratic')
    return interp


def voigt_cdf_numeric(x, gamma, sigma, gridsize=None):
    cdf_x, cdf = voigt_cdf_tab(gamma, sigma, gridsize)
    y = np.interp(x, cdf_x, cdf)
    return y


def voigt_invcdf_numeric(x, gamma, sigma, gridsize=None):
    cdf_x, cdf = voigt_cdf_tab(gamma, sigma, gridsize)
    y = np.interp(x, cdf, cdf_x, left=-np.inf, right=np.inf)
    return y


def voigt_random(gamma, sigma, size, **kwargs):
    """
    Draw random samples from a Voigt distribution.
    
    The tails of the distribution will be clipped;
    the clipping level can be adjusted with the cutoff keyword.
    The default values is 1e-5.
    """
    cdf_x, cdf = voigt_cdf_tab(gamma, sigma, **kwargs)
    random_y = np.random.uniform(np.min(cdf), np.max(cdf), size)
    random_x = np.interp(random_y, cdf, cdf_x)
    return random_x
    



# -*- coding: utf-8 -*-
"""
.. Authors:
    Novimir Pablant <npablant@pppl.gov>

Define the :class:`XicsrtOpticMeshMirror` class.
"""

import numpy as np

from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.optics._InteractMirror import InteractMirror
from xicsrt.optics._ShapeMesh import ShapeMesh

@dochelper
class XicsrtOpticMeshMirror(InteractMirror, ShapeMesh):
    """
    A meshgrid perfect mirror optic.
    """

    pass



# -*- coding: utf-8 -*-
"""
.. Authors:
    Novimir Pablant <npablant@pppl.gov>

Define the :class:`XicsrtOpticMeshToroidalCrystal` class.
"""

import numpy as np

from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.optics._InteractCrystal import InteractCrystal
from xicsrt.optics._ShapeMeshTorus import ShapeMeshTorus

@dochelper
class XicsrtOpticMeshToroidalCrystal(InteractCrystal, ShapeMeshTorus):
    """
    A Toroidal crystal optic implemented using a mesh.

    **Programming Notes**

    The geometry of toroidal mesh was implemented as a separate Shape object
    (:class:`ShapeMeshTorus`) to allow mix-and-match with various Interactions.
    It would have also been possible to simply inherit
    :class:`XicsrtOpticMeshCrystal` and define the geometry here instead. Doing
    so would have avoided the need to create two separate classes and files, but
    would have limit reuse of the defined geometry.
    """

    pass



# -*- coding: utf-8 -*-
"""
.. Authors:
    Novimir Pablant <npablant@pppl.gov>

Define the :class:`XicsrtOpticPlaneMosaicCrystal` class.
"""

import numpy as np

from xicsrt.tools.xicsrt_doc import dochelper
from xicsrt.optics._InteractCrystal import InteractCrystal
from xicsrt.optics._ShapePlane import ShapePlane

@dochelper
class XicsrtOpticPlanarCrystal(InteractCrystal, ShapePlane):
    """
    A planar crystal optic.
    """

    pass



# -*- coding: utf-8 -*-
"""
Code copied from the package fb8 v1.2.2.
https://pypi.org/project/fb8/
https://github.com/tianluyuan/sphere.git

Licence
-------

MIT License
Copyright (c) 2019 Tianlu Yuan

Description
-----------

The algorithms here are partially based on methods described in:
[The Fisher-Bingham Distribution on the Sphere, John T. Kent
Journal of the Royal Statistical Society. Series B (Methodological)
Vol. 44, No. 1 (1982), pp. 71-80 Published by: Wiley
Article Stable URL: http://www.jstor.org/stable/2984712]
"""

from __future__ import print_function
import sys
import warnings
import logging
import heapq

import numpy as np
from scipy.optimize import minimize, basinhopping, approx_fprime
from scipy.special import gamma as G
from scipy.special import gammaln as LG
from scipy.special import iv as I
from scipy.special import ivp as DI
from scipy.special import hyp2f1 as H2F1
from scipy.special import hyp1f1 as H1F1
from scipy.special import hyp0f1 as H0F1
from scipy.integrate import dblquad, IntegrationWarning
import scipy.linalg
from scipy.linalg import eig


# helper function
def MMul(A, B):
    return np.matmul(A, B)


def norm(x, axis=None):
    """
    helper function to compute the L2 norm. scipy.linalg.norm is not used because this function does not allow to choose an axis
    """
    if isinstance(x, list) or isinstance(x, tuple):
        x = np.array(x)
    return np.sqrt(np.sum(x * x, axis=axis))


def fb8(theta, phi, psi, kappa, beta, eta=1., alpha=0., rho=0.):
    """
    Generates the FB8 distribution based on the spherical coordinates theta, phi, psi
    with the concentration parameter kappa and the ovalness beta.
    eta, alpha, and rho set the additional three parameters that allow for asymmetric
    distributions.
    """
    gamma1, gamma2, gamma3 = FB8Distribution.spherical_coordinates_to_gammas(
        theta, phi, psi)
    nu = FB8Distribution.spherical_coordinates_to_nu(alpha, rho)
    fdist = FB8Distribution(gamma1, gamma2, gamma3, kappa, beta, eta, nu)
    if fdist.theta == 0.:
        # if theta == 0, phi and psi are degenerate (rotation in phi == rotation in psi)
        # constructor infers phi as 0 and psi to sum of phi and psi
        fdist.phi = phi
        fdist.psi = psi
    if fdist.alpha == 0.:
        # constructor infers rho as 0 if alpha == 0
        fdist.rho = rho
    return fdist


def fb82(gamma1, gamma2, gamma3, kappa, beta, eta=1., nu=None):
    """
    Generates the FB8 distribution using the orthonormal vectors gamma1,
    gamma2 and gamma3, with the concentration parameter kappa and the ovalness beta.
    eta and nu set the additional three parameters that allow for asymmetric
    distributions.
    """
    assert np.abs(np.inner(gamma1, gamma2)) < 1E-10
    assert np.abs(np.inner(gamma2, gamma3)) < 1E-10
    assert np.abs(np.inner(gamma3, gamma1)) < 1E-10
    return FB8Distribution(gamma1, gamma2, gamma3, kappa, beta, eta, nu)


def fb83(A, B, eta=1., nu=None):
    """
    Generates the FB8 distribution using the orthogonal vectors A and B
    where A = gamma1*kappa and B = gamma2*beta (gamma3 is inferred)
    A may have not have length zero but may be arbitrarily close to zero
    B may have length zero however. If so, then an arbitrary value for gamma2
    (orthogonal to gamma1) is chosen
    """
    kappa = norm(A)
    beta = norm(B)
    gamma1 = A / kappa
    if beta == 0.0:
        gamma2 = __generate_arbitrary_orthogonal_unit_vector(gamma1)
    else:
        gamma2 = B / beta
    theta, phi, psi = FB8Distribution.gammas_to_spherical_coordinates(
        gamma1, gamma2)
    gamma1, gamma2, gamma3 = FB8Distribution.spherical_coordinates_to_gammas(
        theta, phi, psi)
    return FB8Distribution(gamma1, gamma2, gamma3, kappa, beta, eta, nu)


def fb84(Gamma, kappa, beta, eta=1., nu=None):
    """
    Generates the fb8 distribution
    """
    gamma1 = Gamma[:, 0]
    gamma2 = Gamma[:, 1]
    gamma3 = Gamma[:, 2]
    return fb82(gamma1, gamma2, gamma3, kappa, beta, eta, nu)


def __generate_arbitrary_orthogonal_unit_vector(x):
    v1 = np.cross(x, np.array([1.0, 0.0, 0.0]))
    v2 = np.cross(x, np.array([0.0, 1.0, 0.0]))
    v3 = np.cross(x, np.array([0.0, 0.0, 1.0]))
    v1n = norm(v1)
    v2n = norm(v2)
    v3n = norm(v3)
    v = [v1, v2, v3][np.argmax([v1n, v2n, v3n])]
    return v / norm(v)


class FB8Distribution(object):
    minimum_value_for_kappa = 1E-6

    @staticmethod
    def create_matrix_H(theta, phi):
        theta = np.asarray(theta)
        phi = np.asarray(phi)
        _H = np.array([
            [np.cos(theta), -np.sin(theta), np.zeros(theta.shape)],
            [np.sin(theta) * np.cos(phi), np.cos(theta) * np.cos(phi), -np.sin(phi)],
            [np.sin(theta) * np.sin(phi), np.cos(theta) * np.sin(phi), np.cos(phi)]
            ])
        if len(_H.shape) > 2:
            return np.moveaxis(_H, 2, 0)
        else:
            return _H

    @staticmethod
    def create_matrix_Ht(theta, phi):
        return np.swapaxes(FB8Distribution.create_matrix_H(theta, phi), -2, -1)

    @staticmethod
    def create_matrix_K(psi):
        psi = np.asarray(psi)
        zs = np.zeros(psi.shape)
        os = np.ones(psi.shape)
        _K = np.array([
            [os, zs, zs],
            [zs, np.cos(psi), -np.sin(psi)],
            [zs, np.sin(psi), np.cos(psi)]
            ])
        if len(_K.shape) > 2:
            return np.moveaxis(_K, 2, 0)
        else:
            return _K

    @staticmethod
    def create_matrix_Kt(psi):
        return np.swapaxes(FB8Distribution.create_matrix_K(psi), -2, -1)

    @staticmethod
    def create_matrix_Gamma(theta, phi, psi):
        H = FB8Distribution.create_matrix_H(theta, phi)
        K = FB8Distribution.create_matrix_K(psi)
        return MMul(H, K)

    @staticmethod
    def create_matrix_Gammat(theta, phi, psi):
        return np.swapaxes(FB8Distribution.create_matrix_Gamma(theta, phi, psi), -2, -1)

    @staticmethod
    def create_matrix_DH_theta(theta, phi):
        theta = np.asarray(theta)
        phi = np.asarray(phi)
        zs = np.zeros(theta.shape)
        _DH = np.array([
            [-np.sin(theta), -np.cos(theta), zs],
            [np.cos(theta) * np.cos(phi), -np.sin(theta) * np.cos(phi), zs],
            [np.cos(theta) * np.sin(phi), -np.sin(theta) * np.sin(phi), zs]
            ])
        if len(_DH.shape) > 2:
            return np.moveaxis(_DH, 2, 0)
        else:
            return _DH

    @staticmethod
    def create_matrix_DH_phi(theta, phi):
        theta = np.asarray(theta)
        phi = np.asarray(phi)
        zs = np.zeros(theta.shape)
        _DH = np.array([
            [zs, zs, zs],
            [-np.sin(theta) * np.sin(phi), -np.cos(theta) * np.sin(phi), -np.cos(phi)],
            [np.cos(phi) * np.sin(theta), np.cos(theta) * np.cos(phi), -np.sin(phi)]
            ])
        if len(_DH.shape) > 2:
            return np.moveaxis(_DH, 2, 0)
        else:
            return _DH

    @staticmethod
    def create_matrix_DK_psi(psi):
        psi = np.asarray(psi)
        zs = np.zeros(psi.shape)
        _DK = np.array([
            [zs, zs, zs],
            [zs, -np.sin(psi), -np.cos(psi)],
            [zs, np.cos(psi), -np.sin(psi)],
            ])
        if len(_DK.shape) > 2:
            return np.moveaxis(_DK, 2, 0)
        else:
            return _DK

    @staticmethod
    def create_matrix_DGamma_theta(theta, phi, psi):
        return MMul(FB8Distribution.create_matrix_DH_theta(theta, phi), FB8Distribution.create_matrix_K(psi))

    @staticmethod
    def create_matrix_DGamma_phi(theta, phi, psi):
        return MMul(FB8Distribution.create_matrix_DH_phi(theta, phi), FB8Distribution.create_matrix_K(psi))

    @staticmethod
    def create_matrix_DGamma_psi(theta, phi, psi):
        return MMul(FB8Distribution.create_matrix_H(theta, phi), FB8Distribution.create_matrix_DK_psi(psi))

    @staticmethod
    def spherical_coordinates_to_gammas(theta, phi, psi):
        Gamma = FB8Distribution.create_matrix_Gamma(theta, phi, psi)
        gamma1 = Gamma[..., 0]
        gamma2 = Gamma[..., 1]
        gamma3 = Gamma[..., 2]
        return gamma1, gamma2, gamma3

    @staticmethod
    def spherical_coordinates_to_nu(alpha, rho):
        return FB8Distribution.create_matrix_Gamma(
            alpha, rho, np.zeros(np.asarray(alpha).shape))[..., 0]

    @staticmethod
    def gamma1_to_spherical_coordinates(gamma1):
        theta = np.arccos(gamma1[..., 0])
        phi = np.arctan2(gamma1[..., 2], gamma1[..., 1])
        return theta, phi

    @staticmethod
    def gammas_to_spherical_coordinates(gamma1, gamma2):
        theta, phi = FB8Distribution.gamma1_to_spherical_coordinates(gamma1)
        Ht = FB8Distribution.create_matrix_Ht(theta, phi)
        u = MMul(Ht, gamma2.T.reshape(3, np.asarray(theta).size))
        psi = np.arctan2(u[..., 2, :][0], u[..., 1, :][0])
        return theta, phi, psi

    @staticmethod
    def a_c6_star(j, b, k, m):
        assert b > 0
        v = j + 0.5
        return (
            np.exp(
                np.log(b) * j +
                + LG(j + 0.5) - LG(j + 1) - LG(v + 1)
                )
        )  # * H0F1(v+1, k**2/4) * H2F1(-j, 0.5, 0.5-j, -m)

    @staticmethod
    def a_c8_star(jj, kk, ll, b, k, m, n1, n2, n3):
        assert k > 0
        v = jj + ll + kk + 0.5
        z = k * n1
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            ln_n2 = np.log(n2**2) * ll
            ln_n3 = np.log(n3**2) * kk
            ln_b = np.log(b) * jj

        # prevent issues with log for edge cases
        if n2 == 0.:
            ln_n2[ll == 0] = 0
        if n3 == 0.:
            ln_n3[kk == 0] = 0
        if b == 0.:
            ln_b[jj == 0] = 0
        return (
            np.exp(
                ln_n2 + ln_n3 + ln_b +
                np.log(k) * 2 * (ll + kk) -
                LG(2 * ll + 1) - LG(2 * kk + 1) - LG(jj + 1) +
                LG(jj + ll + 0.5) + LG(kk + 0.5) - LG(v + 1) -
                0.5 * np.log(np.pi))  # * H0F1(v+1, z**2/4) * H2F1(-jj, kk+0.5, 0.5-jj-ll, -m)
        )

    def __init__(self, gamma1, gamma2, gamma3, kappa, beta, eta=1., nu=None):
        assert not kappa < 0.
        assert not beta < 0.
        assert not abs(eta) > 1.001
        for gamma in gamma1, gamma2, gamma3:
            assert len(gamma) == 3

        self._gamma1 = np.array(gamma1, dtype=np.float64)
        self._gamma2 = np.array(gamma2, dtype=np.float64)
        self._gamma3 = np.array(gamma3, dtype=np.float64)
        self._kappa = float(kappa)
        self._beta = float(beta)
        # Bingham-Mardia, 4-param, small-circle distribution has eta=-1
        self._eta = float(eta)
        # FB8 param nu
        if nu is None:
            nu = FB8Distribution.spherical_coordinates_to_nu(0, 0)
        self._nu = nu

        self._theta, self._phi, self._psi = FB8Distribution.gammas_to_spherical_coordinates(
            self._gamma1, self._gamma2)
        self._alpha, self._rho = FB8Distribution.gamma1_to_spherical_coordinates(self._nu)

        self._cached_rvs = np.empty((0, 3))
        self._rng = np.random.default_rng()

        # save rvs used to calculated level contours to keep levels self-consistent
        self._level_log_pdf = np.empty((0,))

    @property
    def gamma1(self):
        return self._gamma1

    @property
    def gamma2(self):
        return self._gamma2

    @property
    def gamma3(self):
        return self._gamma3

    @property
    def nu(self):
        return self._nu

    @property
    def kappa(self):
        return self._kappa

    @kappa.setter
    def kappa(self, val):
        self._kappa = val
        self._level_log_pdf = np.empty((0,))
        self._cached_rvs = np.empty((0, 3))

    @property
    def beta(self):
        return self._beta

    @beta.setter
    def beta(self, val):
        self._beta = val
        self._level_log_pdf = np.empty((0,))
        self._cached_rvs = np.empty((0, 3))

    @property
    def eta(self):
        return self._eta

    @eta.setter
    def eta(self, val):
        self._eta = val
        self._level_log_pdf = np.empty((0,))
        self._cached_rvs = np.empty((0, 3))

    @property
    def theta(self):
        return self._theta

    @theta.setter
    def theta(self, val):
        self._theta = np.arccos(np.cos(val))
        self._gamma1, self._gamma2, self._gamma3 = self.Gamma.T
        self._level_log_pdf = np.empty((0,))
        self._cached_rvs = np.empty((0, 3))

    @property
    def phi(self):
        return self._phi

    @phi.setter
    def phi(self, val):
        self._phi = np.arctan2(np.sin(val), np.cos(val))
        self._gamma1, self._gamma2, self._gamma3 = self.Gamma.T
        self._level_log_pdf = np.empty((0,))
        self._cached_rvs = np.empty((0, 3))

    @property
    def psi(self):
        return self._psi

    @psi.setter
    def psi(self, val):
        self._psi = np.arctan2(np.sin(val), np.cos(val))
        self._gamma1, self._gamma2, self._gamma3 = self.Gamma.T
        self._level_log_pdf = np.empty((0,))
        self._cached_rvs = np.empty((0, 3))

    @property
    def alpha(self):
        return self._alpha

    @alpha.setter
    def alpha(self, val):
        self._alpha = np.arccos(np.cos(val))
        self._nu = FB8Distribution.spherical_coordinates_to_nu(
            self._alpha, self._rho)
        self._level_log_pdf = np.empty((0,))
        self._cached_rvs = np.empty((0, 3))

    @property
    def rho(self):
        return self._rho

    @rho.setter
    def rho(self, val):
        self._rho = np.arctan2(np.sin(val), np.cos(val))
        self._nu = FB8Distribution.spherical_coordinates_to_nu(
            self._alpha, self._rho)
        self._level_log_pdf = np.empty((0,))
        self._cached_rvs = np.empty((0, 3))

    @property
    def Gamma(self):
        return self.create_matrix_Gamma(self.theta, self.phi, self.psi)

    @property
    def DGamma_theta(self):
        return self.create_matrix_DGamma_theta(self.theta, self.phi, self.psi)

    @property
    def DGamma_phi(self):
        return self.create_matrix_DGamma_phi(self.theta, self.phi, self.psi)

    @property
    def DGamma_psi(self):
        return self.create_matrix_DGamma_psi(self.theta, self.phi, self.psi)

    @property
    def Dnu_alpha(self):
        return self.create_matrix_DH_theta(self.alpha, self.rho)[..., 0]

    @property
    def Dnu_rho(self):
        return self.create_matrix_DH_phi(self.alpha, self.rho)[..., 0]

    def _nnormalize(self, epsabs=1e-3, epsrel=1e-3):
        """
        Perform numerical integration with dblquad. This function can be used for testing and
        exception handling in self.normalize
        """
        # numerical integration
        k, b, m = self.kappa, self.beta, self.eta
        n1, n2, n3 = self.nu
        return dblquad(
            lambda th, ph:np.sin(th) * \
                          np.exp(k * (n1 * np.cos(th) + n2 * np.sin(th) * np.cos(ph) + n3 * np.sin(th) * np.sin(ph)) + \
                                 b * np.sin(th)**2 * (np.cos(ph)**2 - m * np.sin(ph)**2)),
            0., 2. * np.pi, lambda x:0., lambda x:np.pi,
            epsabs=epsabs, epsrel=epsrel)[0]

    def _normalize(self, cache=dict(), return_num_iterations=False):
        """
        Returns the normalization constant/(2pi) of the FB8 distribution.
        """
        k, b, m = self.kappa, self.beta, self.eta
        n1, n2, n3 = self.nu
        j = 0

        def a_c6(j, b, k, m):
            v = j + 0.5
            return self.a_c6_star(j, b, k, m) * H0F1(v + 1, k**2 / 4) * H2F1(-j, 0.5, 0.5 - j, -m)

        def a_c8(jj, kk, ll, b, k, m, n1, n2, n3):
            v = jj + ll + kk + 0.5
            z = k * n1
            return (self.a_c8_star(jj, kk, ll, b, k, m, n1, n2, n3) *
                    H0F1(v + 1, z**2 / 4) * H2F1(-jj, kk + 0.5, 0.5 - jj - ll, -m))

        def push_coord(val, coord):
            if coord not in inheap:
                heapq.heappush(hq, (val, coord))
                inheap.add(coord)

        if (k, b, m, n1, n2) not in cache:
            result = 0.
            if b == 0. and k == 0.:
                result = 2
            # FB6 or BM4-with-eta
            # This is faster than the full FB8 sum
            elif n1 == 1. or k == 0.:
                # exact solution (vmF)
                if b == 0.0:
                    result = 2 / k * np.sinh(k)
                else:
                    prev_abs_a = 0
                    while True:
                        js = np.arange(j * 100, (j + 1) * 100)
                        a = a_c6(js, b, k, m)
                        evens = js % 2 == 0
                        if np.any(a[evens] < 0):
                            logging.info('a < 0 for even j, masking. This is due to an inaccuracy in H2F1')
                            # hack around H2F1 inaccuracy
                            a[(evens) & (a < 0)] = 0
                        sa = a.sum()
                        abs_sa = np.abs(a).sum()
                        ### DEBUG ###
                        # print j, sa
                        result += sa
                        if np.isnan(result):
                            logging.warning('Series result is nan')
                            raise RuntimeWarning
                        if np.isinf(result):
                            logging.warning('Series result is infinity')
                            raise RuntimeWarning
                        j += 1
                        if abs_sa < np.abs(result) * 1E-12 and abs_sa <= prev_abs_a:
                            break
                        prev_abs_a = abs_sa
            # FB8
            else:
                try:
                    approx_argmax = (0, 0, 0)
                    edge = 0
                    step = 2
                    while edge in approx_argmax and step < 16:
                        step *= 2
                        edge = step**2
                        tjs, tks, tls = np.mgrid[0:edge + 1:step,
                                        0:edge + 1:step,
                                        0:edge + 1:step]

                        a = a_c8(tjs, tks, tls, b, k, m, n1, n2, n3)
                        approx_argmax = np.asarray(
                            np.unravel_index(np.nanargmax(np.abs(a)),
                                             a.shape)) * step
                    amj, amk, aml = approx_argmax
                    result = 0
                    inheap = set([(amj, amk, aml)])
                    hq = []
                    while True:
                        jjs, kks, lls = np.mgrid[
                                        max(0, amj - step):amj + step,
                                        max(0, amk - step):amk + step,
                                        max(0, aml - step):aml + step]
                        a = a_c8(jjs, kks, lls, b, k, m, n1, n2, n3)
                        evens = jjs % 2 == 0
                        if np.any(a[evens] < 0):
                            logging.info('a < 0 for even j, masking. This is due to an inaccuracy in H2F1.')
                        # hack around H2F1 inaccuracy
                        a[(evens) & (a < 0)] = 0
                        sa = a.sum()
                        abs_a = np.abs(a)
                        abs_sa = abs_a.sum()
                        result += sa
                        j += 1
                        if abs_sa < np.abs(result) * 1E-12:
                            break
                        _ = step
                        push_coord(-abs_a[:, :, -2:].sum(), (amj, amk, aml + 2 * step))
                        push_coord(-abs_a[:, -2:, :].sum(), (amj, amk + 2 * step, aml))
                        push_coord(-abs_a[-2:, :, :].sum(), (amj + 2 * step, amk, aml))
                        if 0 < aml - step:
                            push_coord(-abs_a[:, :, :2].sum(), (amj, amk, aml - 2 * step))
                        if 0 < amk - step:
                            push_coord(-abs_a[:, :2, :].sum(), (amj, amk - 2 * step, aml))
                        if 0 < amj - step:
                            push_coord(-abs_a[:2, :, :].sum(), (amj - 2 * step, amk, aml))
                        amj, amk, aml = heapq.heappop(hq)[1]

                    if not result > 0:
                        logging.warning('Series result not positive')
                        raise RuntimeWarning

                except (RuntimeWarning, OverflowError) as e:
                    logging.warning(
                        'Series calculation of normalization failed. Attempting numerical integration... ' + self.__repr__())
                    try:
                        # numerical integration
                        result = self._nnormalize() / (2 * np.pi)
                    except RuntimeWarning as e:
                        result = np.inf
                    j = -1

            cache[k, b, m, n1, n2] = result

        if return_num_iterations:
            return cache[k, b, m, n1, n2], j
        else:
            return cache[k, b, m, n1, n2]

    def normalize(self, return_num_iterations=False):
        """
        Returns the normalization constant of the FB8 distribution.
        The proportional error may be expected not to be greater than
        1E-11.
        >>> gamma1 = np.array([1.0, 0.0, 0.0])
        >>> gamma2 = np.array([0.0, 1.0, 0.0])
        >>> gamma3 = np.array([0.0, 0.0, 1.0])
        >>> tiny = FB8Distribution.minimum_value_for_kappa
        >>> np.abs(fb82(gamma1, gamma2, gamma3, tiny, 0.0).normalize() - 4*np.pi) < 4*np.pi*1E-12
        True
        >>> for kappa in [0.01, 0.1, 0.2, 0.5, 2, 4, 8, 16]:
        ...     print(np.abs(fb82(gamma1, gamma2, gamma3, kappa, 0.0).normalize() - 4*np.pi*np.sinh(kappa)/kappa) < 1E-15*4*np.pi*np.sinh(kappa)/kappa, end=' ')
        ...
        True True True True True True True True
        """
        if return_num_iterations:
            res, nit = self._normalize(return_num_iterations=return_num_iterations)
            return 2 * np.pi * res, nit
        return 2 * np.pi * self._normalize(return_num_iterations=return_num_iterations)

    def _approx_log_normalize(self):
        """
        >>> from itertools import product
        >>> for x in product([0], [0], [0], range(10, 200, 10),
        ...                 range(10, 200, 10),
        ...                 np.linspace(-1, 1, 5), [0], [0]):
        ...    lnorm = np.log(fb8(*x).normalize())
        ...    lnormapprox = fb8(*x)._approx_log_normalize()
        ...    if np.abs(lnorm-lnormapprox)/lnorm > 0.1:
        ...        print(fb8(*x), lnorm, lnormapprox)
        """
        # should only reach here if FB6 or BM4-with-eta
        assert self.nu[0] == 1 or self.kappa == 0
        k = self.kappa
        b = self.beta
        m = self.eta
        if k > 2 * b:
            lnormalize = np.log(2 * np.pi) + k - \
                         np.log((k - 2 * b) * (k + 2 * b * m)) / 2.
        else:
            # normal approximation in z = cos(theta) with correction factor for floating eta
            # correction factor by fixing sin(theta)**2 = (1-k**2/(4*b**2)), corresponding to theta_max
            # Written in terms of 1F1 using A&S (13.1.27) in which
            # I(0, z) = M(1/2, 1, -2z) exp(z)
            _ = k**2 / (4 * b**2)
            z = (1 + m) * b * (1 - _) / 2
            lnormalize = ((np.log(np.pi) - np.log(b)) / 2 + b * (1 + _) +
                          np.log(2 * np.pi) + np.log(H1F1(1 / 2., 1., -2 * z)))

        return lnormalize

    def log_normalize(self):
        """
        Returns the logarithm of the normalization constant.
        >>> print('Overflow edgecase:', fb8(0.67, 1.02, 0.39, 115.42, 615.17, -0.92, 1.15, 0.49).log_normalize())
        Overflow edgecase: 710.9592764244503
        >>> from itertools import product
        >>> for x in product([0], [0], [0], [0, 2, 32, 128, 256],
        ...                  [0, 2, 32, 128, 256], np.linspace(-1, 1, 5),
        ...                  np.linspace(0.0, np.pi, 3),
        ...                  np.linspace(0, np.pi/3, 3)):
        ...    lnorm = fb8(*x).log_normalize()
        ...    lnnorm = np.log(fb8(*x)._nnormalize())
        ...    if np.abs(lnorm-lnnorm)/lnorm > 0.1:
        ...        print(fb8(*x), round(lnorm, 5), round(lnnorm, 5))
        fb8(0.00, 0.00, 0.00, 256.00, 128.00, 1.00, 1.57, 1.05) 337.92902 289.82614
        fb8(0.00, 0.00, 0.00, 256.00, 256.00, 1.00, 1.57, 1.05) 466.23216 400.40835
        """
        with warnings.catch_warnings():
            warnings.simplefilter('error')
            try:
                return np.log(self._normalize()) + np.log(2 * np.pi)
            except (OverflowError, RuntimeWarning) as e:
                logging.warning(
                    'Series calculation of normalization failed. Approximating normalization... ' + self.__repr__())
                return self._approx_log_normalize()

    def _grad_log_normalize(self, cache=dict(), return_num_iterations=False):
        """ Derivative of the log-normalization constant wrt k, b, m, alpha, rho
        >>> def func(x):
        ...     return fb8(0,0,0,*x).log_normalize()
        >>> def grad(x):
        ...     return fb8(0,0,0,*x)._grad_log_normalize()
        >>> from scipy.optimize import check_grad
        >>> from itertools import product
        >>> for x in product([0.0, 16, 32],
        ...                  [0.0, 16, 32], np.linspace(-0.99, 0.99, 3)+1e-5,
        ...                  np.linspace(0, np.pi-1e-3, 3),
        ...                  np.linspace(0, np.pi/3-1e-3, 3)):
        ...     if check_grad(func, grad, x) > 1:
        ...         print(fb8(0,0,0,*x), check_grad(func, grad, x))
        """
        k, b, m = self.kappa, self.beta, self.eta
        n1, n2, n3 = self.nu
        alpha, rho = self.alpha, self.rho
        eps = 1e-6
        if k == 0.:
            k = eps
        if b == 0.:
            b = eps
        if m == -1.:
            m = -1 + eps
        elif m == 1.:
            m = 1 - eps
        if n2 == 0.:
            n2 = eps
        if n3 == 0.:
            n3 = eps

        def grad_a_c6(j, b, k, m):
            v = j + 0.5
            h0f1_c6 = H0F1(v + 1, k**2 / 4)
            h2f1_c6 = H2F1(-j, 0.5, 0.5 - j, -m)
            v = j + 0.5
            a_c6_st = self.a_c6_star(j, b, k, m)
            _Da_b = j / b * a_c6_st * h0f1_c6 * h2f1_c6
            _Da_k = H0F1(v + 2, k**2 / 4) * k / (2 * (1 + v)) * a_c6_st * h2f1_c6
            _Da_m = 0.5 * j * H2F1(1 - j, 1.5, 1.5 - j, -m) / (0.5 - j) * a_c6_st * h0f1_c6
            return _Da_k, _Da_b, _Da_m

        def grad_a_c8(jj, kk, ll, b, k, m, n1, n2, n3):
            v = jj + ll + kk + 0.5
            z = k * n1
            a_c8_st = self.a_c8_star(jj, kk, ll, b, k, m, n1, n2, n3)
            h0f1_c8 = H0F1(v + 1, z**2 / 4)
            dh0f1_c8 = H0F1(v + 2, z**2 / 4) / (2 * (v + 1))
            h2f1_c8 = H2F1(-jj, kk + 0.5, 0.5 - jj - ll, -m)
            # dh2f1_c8 = np.zeros(h2f1_c8.shape)
            # dh2f1_c8[:-1,:-1,:-1] = h2f1_c8[1:,1:,1:]
            # dh2f1_c8[-1,...] = H2F1(1-jj[-1,...], 1.5+kk[-1,...], 1.5-jj[-1,...]-ll[-1,...], -m)
            # dh2f1_c8[:,-1,:] = H2F1(1-jj[:,-1,:], 1.5+kk[:,-1,:], 1.5-jj[:,-1,:]-ll[:,-1,:], -m)
            # dh2f1_c8[...,-1] = H2F1(1-jj[...,-1], 1.5+kk[...,-1], 1.5-jj[...,-1]-ll[...,-1], -m)
            hprd_c8 = h0f1_c8 * h2f1_c8

            _Da_b = jj / b * a_c8_st * hprd_c8
            _Da_k = (2 / k * (kk + ll) * h0f1_c8 + k * n1**2 * dh0f1_c8) * a_c8_st * h2f1_c8
            _Da_m = jj * (kk + 0.5) / (0.5 - jj - ll) * H2F1(1 - jj, 1.5 + kk, 1.5 - jj - ll, -m) * a_c8_st * h0f1_c8
            # if np.any(np.isnan(_Da_m)):
            #     import pdb
            #     pdb.set_trace()
            _Da_n1 = k**2 * n1 * dh0f1_c8 * a_c8_st * h2f1_c8
            _Da_n2 = 2 * ll / n2 * a_c8_st * hprd_c8
            _Da_n3 = 2 * kk / n3 * a_c8_st * hprd_c8
            _Da_nu = np.asarray([_Da_n1, _Da_n2, _Da_n3])

            return _Da_k, _Da_b, _Da_m, np.tensordot(self.Dnu_alpha, _Da_nu, 1), np.tensordot(self.Dnu_rho, _Da_nu, 1)

        if (k, b, m, n1, n2) not in cache:
            snorm = 2 * np.pi / np.exp(self.log_normalize())
            j = 0
            result = np.zeros([5, ])
            # FB6
            # This is faster than the full FB8 sum
            if n1 == 1.:
                prev_abs_a = 0
                while True:
                    js = np.arange(j * 100, (j + 1) * 100)
                    grad_a = np.asarray(grad_a_c6(js, b, k, m))
                    sa = grad_a.sum(axis=1) * snorm
                    abs_sa = np.abs(grad_a).sum(axis=1) * snorm
                    result[:3] += sa
                    if np.any(np.isnan(result)) or np.any(np.isinf(result)):
                        logging.warning(
                            'Series grad(ln(c6)) is nan or infinity, using approx_fprime...' + self.__repr__())
                        result[:3] = approx_fprime((k, b, m), lambda x:fb8(0, 0, 0, *x).log_normalize(),
                                                   1.49e-8)
                        j = -1
                        break
                    j += 1
                    if np.all(abs_sa <= np.abs(result[:3]) * 1E-3) and np.all(abs_sa <= prev_abs_a):
                        break
                    prev_abs_a = abs_sa
            # FB8
            else:
                ll = 0
                prev_abs_sa_ll = 0
                _l, _k, _j = (14,) * 3
                _jjs, _kks, _lls = np.mgrid[0:_j, 0:_k, 0:_l]
                while True:
                    lls = ll * _l + _lls
                    curr_abs_sa_ll = 0
                    kk = 0
                    prev_abs_sa_kk = 0
                    while True:
                        kks = kk * _k + _kks
                        curr_abs_sa_kk = 0
                        jj = 0
                        prev_abs_sa_jj = 0
                        while True:
                            jjs = jj * _j + _jjs
                            grad_a = np.asarray(grad_a_c8(jjs, kks, lls, b, k, m, n1, n2, n3))
                            sa = grad_a.sum(axis=(1, 2, 3)) * snorm
                            abs_sa = np.abs(grad_a).sum(axis=(1, 2, 3)) * snorm
                            ### DEBUG ###
                            # import pdb
                            # pdb.set_trace()
                            # print ll, kk, jj, sa, abs_sa
                            # print j, a, I(j+0.5, k)
                            # print(j,ll,kk,jj, result*2*np.pi/norm)
                            if np.any(np.isnan(sa)):
                                logging.warning(
                                    'Series grad(ln(c_8)) is nan, using approx_fprime...' + self.__repr__())
                                result = approx_fprime((k, b, m, alpha, rho), lambda x:fb8(0, 0, 0, *x).log_normalize(),
                                                       1.49e-8)
                                j = -1
                                break
                            curr_abs_sa_kk += abs_sa
                            curr_abs_sa_ll += abs_sa
                            result += sa
                            j += 1
                            jj += 1
                            if np.all(abs_sa <= np.abs(result) * 1E-3) and np.all(abs_sa <= prev_abs_sa_jj):
                                break
                            prev_abs_sa_jj = abs_sa
                        kk += 1
                        if (j == -1) or (np.all(curr_abs_sa_kk <= np.abs(result) * 1E-3) and
                                         np.all(curr_abs_sa_kk <= prev_abs_sa_kk)):
                            break
                        prev_abs_sa_kk = curr_abs_sa_kk
                    ll += 1
                    if (j == -1) or (np.all(curr_abs_sa_ll <= np.abs(result) * 1E-3) and
                                     np.all(curr_abs_sa_ll <= prev_abs_sa_ll)):
                        # print(jj, kk, ll, j)
                        break
                    prev_abs_sa_ll = curr_abs_sa_ll

            cache[k, b, m, n1, n2] = result

        if return_num_iterations:
            return cache[k, b, m, n1, n2], j
        else:
            return cache[k, b, m, n1, n2]

    def max(self):
        k, b, m = self.kappa, self.beta, self.eta
        n1, n2, n3 = self.nu
        if n1 == 1.:
            if self.beta == 0.0:
                x1 = 1
            else:
                x1 = self.kappa / (2 * self.beta)
            if x1 > 1:
                x1 = 1
            x2 = np.sqrt(1 - x1**2)
            x3 = 0
        else:
            # FB8
            # Line search along solution set with BFGS run over x_max
            ntests = 1001
            radicalz = lambda z:4 * m**2 * z**2 * (z**2 - 1) * b**2 + \
                                4 * m * z * (z**2 - 1) * b * k * n1 + k**2 * ((z**2 - 1) * n1**2 + z**2 * n3**2)

            curr_max = -np.inf
            x_max = None
            for sgn_0 in [-1, 1]:
                for sgn_1 in [-1, 1]:
                    x1 = np.linspace(-1, 1, ntests)
                    x2 = sgn_0 * np.sqrt(-radicalz(x1)) / (2 * m * x1 * b + k * n1)
                    x3 = sgn_1 * np.sqrt(1 - x1**2 - x2**2)

                    x = np.asarray((x1, x2, x3))
                    lpdfs = self.log_pdf(np.dot(self.Gamma, x).T, normalize=False)
                    lpdfs_max = np.nanmax(lpdfs)
                    # print lpdfs_max
                    if lpdfs_max > curr_max:
                        x_max = x.T[np.nanargmax(lpdfs)]
                        curr_max = lpdfs_max

            f = lambda x:-k * (
                    n1 * np.cos(x[0]) + n2 * np.sin(x[0]) * np.cos(x[1]) + n3 * np.sin(x[0]) * np.sin(x[1])) - \
                         b * np.sin(x[0])**2 * (np.cos(x[1])**2 - m * np.sin(x[1])**2)
            _x = minimize(f, self.gamma1_to_spherical_coordinates(x_max))
            if not _x.success:
                warning_message = _x.message
                warnings.warn(warning_message, RuntimeWarning)
            x1, x2, x3 = self.spherical_coordinates_to_nu(*_x.x)

        x = np.dot(self.Gamma, np.asarray((x1, x2, x3)))
        return FB8Distribution.gamma1_to_spherical_coordinates(x)

    def pdf_max(self, normalize=True):
        return np.exp(self.log_pdf_max(normalize))

    def log_pdf_max(self, normalize=True):
        """
        Returns the maximum value of the log(pdf)
        """
        return self.log_pdf(FB8Distribution.spherical_coordinates_to_nu(
            *self.max()), normalize)

    def pdf(self, xs, normalize=True):
        """
        Returns the pdf of the fb8 distribution for 3D vectors that
        are stored in xs which must be an array of N x 3 or N x M x 3
        N x M x P x 3 etc.
        The code below shows how points in the pdf can be evaluated. An integral is
        calculated using random points on the sphere to determine wether the pdf is
        properly normalized.
        >>> from numpy.random import default_rng
        >>> rng = default_rng(666)
        >>> num_samples = 400000
        >>> xs = rng.normal(0, 1, (num_samples,3))
        >>> xs = np.divide(xs, np.reshape(norm(xs, 1), (num_samples, 1)))
        >>> assert np.abs(4*np.pi*np.average(fb8(1.0, 1.0, 1.0, 4.0,  2.0).pdf(xs)) - 1.0) < 0.01
        >>> assert np.abs(4*np.pi*np.average(fb8(1.0, 2.0, 3.0, 4.0,  2.0).pdf(xs)) - 1.0) < 0.01
        >>> assert np.abs(4*np.pi*np.average(fb8(1.0, 2.0, 3.0, 4.0,  8.0).pdf(xs)) - 1.0) < 0.01
        >>> assert np.abs(4*np.pi*np.average(fb8(1.0, 2.0, 3.0, 16.0, 8.0).pdf(xs)) - 1.0) < 0.01
        """
        return np.exp(self.log_pdf(xs, normalize))

    def log_pdf(self, xs, normalize=True):
        """
        Returns the log(pdf) of the fb8 distribution.
        """
        g1x, g2x, g3x = MMul(self.Gamma.T, np.asarray(xs).T)
        k, b, m = self.kappa, self.beta, self.eta
        ngx = self.nu.dot(np.asarray([g1x, g2x, g3x]))

        f = k * ngx + b * (g2x**2 - m * g3x**2)
        if normalize:
            return f - self.log_normalize()
        else:
            return f

    def _grad_log_pdf(self, xs):
        """
        Returns the gradient of the log(pdf(xs)) over the parameters.
        """
        gx = MMul(self.Gamma.T, np.asarray(xs).T)
        dgx_theta = MMul(self.DGamma_theta.T, np.asarray(xs).T)
        dgx_phi = MMul(self.DGamma_phi.T, np.asarray(xs).T)
        dgx_psi = MMul(self.DGamma_psi.T, np.asarray(xs).T)
        k, b, m = self.kappa, self.beta, self.eta
        ngx = self.nu.dot(gx)

        # f = k * ngx + b * (g2x**2 - m * g3x**2)
        Df_k = ngx
        Df_b = gx[1]**2 - m * gx[2]**2
        Df_m = -b * gx[2]**2
        Df_theta = k * self.nu.dot(dgx_theta) + 2 * b * (gx[1] * dgx_theta[1] - m * gx[2] * dgx_theta[2])
        Df_phi = k * self.nu.dot(dgx_phi) + 2 * b * (gx[1] * dgx_phi[1] - m * gx[2] * dgx_phi[2])
        Df_psi = k * self.nu.dot(dgx_psi) + 2 * b * (gx[1] * dgx_psi[1] - m * gx[2] * dgx_psi[2])
        Df_alpha = k * self.Dnu_alpha.dot(gx)
        Df_rho = k * self.Dnu_rho.dot(gx)
        _ = self._grad_log_normalize()
        # print(Df_b, _[1])
        return Df_theta, Df_phi, Df_psi, Df_k - _[0], Df_b - _[1], Df_m - _[2], Df_alpha - _[3], Df_rho - _[4]

    def log_likelihood(self, xs):
        """
        Returns the log likelihood for xs.
        """
        retval = self.log_pdf(xs)
        return sum(retval, len(np.shape(retval)) - 1)

    def grad_log_likelihood(self, xs):
        """
        Returns the gradient of the log likelihood given xs over all 8 parameters.
        >>> def func_llh(x, xs):
        ...     return fb8(*x).log_likelihood(xs)
        >>> def grad_llh(x, xs):
        ...     return fb8(*x).grad_log_likelihood(xs)
        >>> from scipy.optimize import check_grad
        >>> from itertools import product
        >>> xs = np.array([[ 0.72692034, -0.58196172,  0.36456465],
        ...                [ 0.58726806,  0.25163898, -0.76928152],
        ...                [ 0.35595372,  0.77330355,  0.52468902]])
        >>> for x in product([0,1], [-1,0,1], [-1, 0,1], [0.0, 2, 32],
        ...                  [0.0, 2, 32], np.linspace(-0.99, 0.99, 3),
        ...                  np.linspace(0, np.pi-1e-3, 2),
        ...                  np.linspace(0, np.pi/3-1e-3, 2)):
        ...     if check_grad(func_llh, grad_llh, x, xs) > 1:
        ...         print(x, check_grad(func_llh, grad_llh, x, xs))
        """
        gradval = self._grad_log_pdf(xs)
        return [sum(_, len(np.shape(_)) - 1) for _ in gradval]

    def _rvs_helper(self):
        num_samples = 10000
        xs = self._rng.normal(0, 1, (num_samples, 3))
        xs = np.divide(xs, np.reshape(norm(xs, 1), (num_samples, 1)))
        lpvalues = self.log_pdf(xs, normalize=False)
        lfmax = self.log_pdf_max(normalize=False)
        ## DEBUG
        # print lfmax, lpvalues.max()
        # assert lfmax > lpvalues.max()
        ## END
        shifted = lpvalues - lfmax
        return xs[self._rng.uniform(0, 1, num_samples) < np.exp(shifted)]

    def rvs(self, n_samples=None, seed=False):
        """
        Returns random samples from the FB8 distribution by rejection sampling.
        May become inefficient for large kappas.
        The returned random samples are 3D unit vectors.
        If n_samples == None then a single sample x is returned with shape (3,)
        If n_samples is an integer value N then N samples are returned in an array with shape (N, 3)
        If seed is None, int, array_like[ints], SeedSequence, BitGenerator, or Generator, the cache
        is reset and fresh random variates will be drawn using numpy.random.default_rng(seed).
        If seed is False, the cache will be used, in conjunction with the last rng state if needed.
        >>> k = fb8(1.0, 2.0, 3.0, 16.0, 8.0)
        >>> assert np.sum(k.rvs(100, 1)-k.rvs(100, 1)) == 0.
        >>> assert np.sum(k.rvs(100, 1)-k.rvs(200, 1)[:100]) == 0.
        >>> assert np.all(k.rvs(100)-k.rvs(100, None))
        >>> assert len(np.unique(k.rvs(10000, 1)[:,0])) == 10000
        """
        num_samples = 1 if n_samples is None else n_samples
        if seed is not False:
            self._rng = np.random.default_rng(seed)
            self._cached_rvs = np.empty((0, 3))
        rvs = self._cached_rvs
        while len(rvs) < num_samples:
            new_rvs = self._rvs_helper()
            rvs = np.concatenate([rvs, new_rvs])
        if n_samples == None:
            self._cached_rvs = rvs[1:]
            return rvs[0]
        else:
            self._cached_rvs = rvs[num_samples:]
            retval = rvs[:num_samples]
            return retval

    def level(self, percentile=50, n_samples=10000, seed=None):
        """
        Returns the -log_pdf level at percentile by generating a set of rvs and their log_pdfs
        """
        if 0 <= percentile < 100:
            curr_len = self._level_log_pdf.size
            if curr_len < n_samples:
                new_rvs = self.rvs(n_samples, seed)
                self._level_log_pdf = -self.log_pdf(new_rvs)
                self._level_log_pdf.sort()
            log_pdf = self._level_log_pdf
            loc = (log_pdf.size - 1) * percentile / 100.
            idx, frac = int(loc), loc - int(loc)
            return log_pdf[idx] + frac * (log_pdf[idx + 1] - log_pdf[idx])
        else:
            print('{} percentile out of bounds'.format(percentile))
            return np.nan

    def contour(self, percentile=50):
        """
        Returns the (spherical) coordinates that correspond to a contour percentile
        Solution is based on Eq 1.4 (Kent 1982)
        """
        k = self.kappa
        b = self.beta
        m = self.eta
        ln = self.log_normalize()
        lev = self.level(percentile)

        # FB6, exact
        if self.nu[0] == 1.:
            # work in coordinate system x' = Gamma'*x
            # range over which x1 remains real is [-x2_max, x2_max]
            # assume -1 < m < 1
            x2_max = min(np.abs(np.sqrt(k**2 + 4 * b * m * (b * m - lev + ln)
                                        ) / (2 * np.sqrt(m * (1 + m)) * b)), 1)
            if np.isnan(x2_max):
                x2_max = 1
            x2 = np.linspace(-x2_max, x2_max, 10000)
            if m == 0. or b == 0.:
                x1_0 = -(lev - ln + b * x2**2) / k
                x1_1 = x1_0
            else:
                x1_0 = (-k + np.sqrt(k**2 + 4 * m * b * (m * b - lev +
                                                         ln - (1 + m) * b * x2**2))) / (2 * b * m)
                x1_1 = (-k - np.sqrt(k**2 + 4 * m * b * (m * b - lev +
                                                         ln - (1 + m) * b * x2**2))) / (2 * b * m)
            x3_0 = np.sqrt(1 - x1_0**2 - x2**2)
            x3_1 = np.sqrt(1 - x1_1**2 - x2**2)
            x2 = np.concatenate([x2, -x2, x2, -x2])
            x1 = np.concatenate([x1_0, x1_0, x1_1, x1_1])
            x3 = np.concatenate([x3_0, -x3_0, x3_1, -x3_1])

            # Since Kent distribution is well-defined for points not on a sphere,
            # possible solutions for L=-log_pdf(kent) extend beyond surface of
            # sphere. For the contour evaluation, only use points that lie on sphere.
            ok = x1**2 + x2**2 <= 1
            x123 = np.asarray((x1[ok], x2[ok], x3[ok]))

            # rotate back into x coordinates
            x = np.dot(self.Gamma, x123).T
        # FB8 approximate
        else:
            npts = 10000
            rvs = self.rvs(npts)
            x = rvs[np.argsort(np.abs(lev + self.log_pdf(rvs)))[:200]]
        return FB8Distribution.gamma1_to_spherical_coordinates(x)

    def __repr__(self):
        return f'fb8({self.theta:.2f}, {self.phi:.2f}, {self.psi:.2f}, {self.kappa:.2f}, {self.beta:.2f}, {self.eta:.2f}, {self.alpha:.2f}, {self.rho:.2f})'


def kent_me(xs):
    """Generates and returns a FB8Distribution based on a FB5 (Kent) moment estimation."""
    lenxs = len(xs)
    xbar = np.average(xs, 0)  # average direction of samples from origin
    # dispersion (or covariance) matrix around origin
    S = np.average(xs.reshape((lenxs, 3, 1)) * xs.reshape((lenxs, 1, 3)), 0)
    # has unit length and is in the same direction and parallel to xbar
    gamma1 = xbar / norm(xbar)
    theta, phi = FB8Distribution.gamma1_to_spherical_coordinates(gamma1)

    H = FB8Distribution.create_matrix_H(theta, phi)
    Ht = FB8Distribution.create_matrix_Ht(theta, phi)
    B = MMul(Ht, MMul(S, H))

    eigvals, eigvects = eig(B[1:, 1:])
    eigvals = np.real(eigvals)
    if eigvals[0] < eigvals[1]:
        eigvals[0], eigvals[1] = eigvals[1], eigvals[0]
        eigvects = eigvects[:, ::-1]
    K = np.diag([1.0, 1.0, 1.0])
    K[1:, 1:] = eigvects

    G = MMul(H, K)
    Gt = np.swapaxes(G, -2, -1)
    T = MMul(Gt, MMul(S, G))

    r1 = norm(xbar)
    t22, t33 = T[1, 1], T[2, 2]
    r2 = t22 - t33

    # kappa and beta can be estimated but may not lie outside their permitted ranges
    min_kappa = FB8Distribution.minimum_value_for_kappa
    kappa = max(min_kappa, 1.0 / (2.0 - 2.0 * r1 - r2) +
                1.0 / (2.0 - 2.0 * r1 + r2))
    beta = 0.5 * (1.0 / (2.0 - 2.0 * r1 - r2) - 1.0 / (2.0 - 2.0 * r1 + r2))

    return fb84(G, kappa, beta)


def __fb8_mle_output1(k_me, callback):
    print()
    print("******** Maximum Likelihood Estimation ********")
    print("Initial estimates are:")
    print("theta =", k_me.theta)
    print("phi   =", k_me.phi)
    print("psi   =", k_me.psi)
    print("kappa =", k_me.kappa)
    print("beta  =", k_me.beta)
    print("eta   =", k_me.eta)
    print("alpha =", k_me.alpha)
    print("rho   =", k_me.rho)
    print("******** Starting the Gradient Descent ********")
    print("[iteration]   fb8(theta, phi, psi, kappa, beta, eta, alpha, rho)   -L")


def fb8_mle(xs, verbose=False, return_intermediate_values=False, warning='warn', fb5_only=False):
    """
    Generates a FB8Distribution fitted to xs using maximum likelihood estimation
    For a first approximation kent_me() is used. The function
    -k.log_likelihood(xs)/len(xs) (where k is an instance of FB8Distribution) is
    minimized. If fb5_only=False, sequentially fit a Kent, FB6 and then FB8
    distribution. Gradients are used for the FB6 and FB8 fits.
    Input:
      - xs: values on the sphere to be fitted by MLE, ordering is (z, x, y)
      - verbose: if True, output is given for every step
      - return_intermediate_values: if true the values of all intermediate steps
        are returned as well
      - warning: choices are
        - "warn": issues any warning via warning.warn
        - a file object: which results in any warning message being written to a file
          (e.g. stdout)
        - "none": or any other value for this argument results in no warnings to be issued
      - fb5_only: perform fit to Kent distribution only
    Output:
      - an instance of the fitted FB8Distribution
    Extra output:
      - if return_intermediate_values is specified then
      a tuple is returned with the FB8Distribution argument as the first element
      and containing the extra requested values in the rest of the elements.
    """
    lenxs = len(xs)

    # method that generates the minus L to be minimized
    # x = theta phi psi kappa beta eta alpha rho
    def minus_log_likelihood(x):
        if np.any(np.isnan(x)):
            return np.inf
        if x[3] < 0 or x[4] < 0:
            return np.inf
        ### DEBUG ###
        # if len(x) > 5 and (x[5] > 1 or x[5] < -1):
        #     return np.inf
        return -fb8(*x).log_likelihood(xs) / lenxs

    def jac(x):
        if np.any(np.isnan(x)):
            return np.zeros(8)
        if x[3] < 0 or x[4] < 0:
            return np.zeros(8)
        return -np.asarray(fb8(*x).grad_log_likelihood(xs)[:len(x)]) / lenxs

    # callback for keeping track of the values
    intermediate_values = list()

    def callback(x, output_count=[0]):
        kx = fb8(*x)
        minusL = -kx.log_likelihood(xs)
        imv = intermediate_values
        imv.append((x, minusL))
        if verbose:
            print(len(imv), kx, minusL)

    # first get estimated moments
    k_me = kent_me(xs)
    theta, phi, psi, kappa, beta = k_me.theta, k_me.phi, k_me.psi, k_me.kappa, k_me.beta

    # here the mle is done
    x_start = np.array([theta, phi, psi, kappa, beta])
    y_start = np.array([theta, phi, psi, beta, kappa, -0.99])

    # First try a FB5 fit
    # constrain kappa, beta >= 0 and 2*beta <= kappa for FB5 (Kent 1982)
    if verbose:
        __fb8_mle_output1(fb8(*x_start), callback)
    cons = ({"type":"ineq",
             "fun":lambda x:x[3] - 2 * x[4]},
            {"type":"ineq",
             "fun":lambda x:x[3]},
            {"type":"ineq",
             "fun":lambda x:x[4]})
    all_values = minimize(minus_log_likelihood,
                          x_start,
                          method="SLSQP",
                          constraints=cons,
                          callback=callback,
                          options={"disp":False, "ftol":1e-09,
                                   "maxiter":100})

    if not fb5_only:
        # Then try a FB6 fit with seed: eta = -0.99
        # note eta=-1 with 2*beta >= kappa is the small-circle distribution (Bingham-Mardia 1978)
        if verbose:
            __fb8_mle_output1(fb8(*y_start), callback)
        ### constraint for SLSQP ###
        # cons = ({"type": "ineq", # kappa >= 0
        #          "fun": lambda x: x[3]},
        #         {"type": "ineq", # beta >= 0
        #          "fun": lambda x: x[4]},
        #         {"type": "ineq", # -1 <= eta <=1
        #          "fun": lambda x: 1 - np.abs(x[5])})
        #         # {"type": "ineq",
        #         #  "fun": lambda x: -x[3] + 2 * x[4]})
        lb = [0., -np.pi, -np.pi, 0, 0, -1, 0.01, -np.pi]
        ub = [np.pi, np.pi, np.pi, None, None, 1, np.pi, np.pi]
        _y = minimize(minus_log_likelihood,
                      y_start,
                      jac=jac,
                      method="L-BFGS-B",
                      bounds=list(zip(lb[:6], ub[:6])),
                      callback=callback)

        # Choose better of FB5 vs FB6 as another seed for FB8
        # Last three parameters determine if FB5, FB6, or FB8
        z_starts = [np.array([np.abs(theta - np.pi / 2), phi, psi, beta, kappa, -0.9, np.pi / 4, 0.]), ]
        if _y.success and _y.fun < all_values.fun:
            all_values = _y
            z_starts.append(np.concatenate((_y.x, [0.2, 0.])))
        else:
            z_starts.append(np.concatenate((all_values.x, [0.9, 0.2, 0.])))

        for z_start in z_starts:
            if verbose:
                __fb8_mle_output1(fb8(*z_start), callback)
            # _z = basinhopping(minus_log_likelihood,
            #                   z_start,
            #                   niter=10,
            #                   niter_success=3,
            #                   minimizer_kwargs= dict(
            #                       method="L-BFGS-B",
            #                       callback=callback,
            #                       options={"disp": False,
            #                                "maxiter": 100})).lowest_optimization_result
            _z = minimize(minus_log_likelihood,
                          z_start,
                          jac=jac,
                          method="L-BFGS-B",
                          bounds=list(zip(lb, ub)),
                          callback=callback,
                          options={'ftol':1e-8, 'gtol':1e-4})
            if _z.success and _z.fun < all_values.fun:
                all_values = _z
    if not all_values.success:
        warning_message = all_values.message
        if warning == "warn":
            warnings.warn(warning_message, RuntimeWarning)
        if hasattr(warning, "write"):
            warning.write("Warning: " + warning_message + "\n")

    k = (fb8(*all_values.x),)
    if return_intermediate_values:
        k += (intermediate_values,)
    if len(k) == 1:
        k = k[0]
    return k


if __name__ == "__main__":
    __doc__ += """
>>> import numpy as np
>>> from sphere.example import test_example_normalization, test_example_mle, test_example_mle2
>>> test_example_normalization(gridsize=10)
Calculating the matrix M_ij of values that can be calculated: kappa=100.0*i+1, beta=100.0*j+1
with eta=1.0, alpha=0.0, rho=0.0
Calculating normalization factor for combinations of kappa and beta:
Iterations necessary to calculate normalize(kappa, beta):
  2   3   5   6   7   8   9  10   x   x
  2   3   4   6   7   8   9  10   x   x
  2   3   4   6   7   8   9   x   x   x
  2   2   4   5   7   8   9   x   x   x
  2   2   3   5   6   7   9   x   x   x
  2   2   2   4   6   7   8   x   x   x
  2   2   2   3   5   7   x   x   x   x
  2   2   2   3   4   x   x   x   x   x
  x   x   x   x   x   x   x   x   x   x
  x   x   x   x   x   x   x   x   x   x
>>> logging.getLogger().setLevel('ERROR')
>>> test_example_normalization(gridsize=10,alpha=0.5)
Calculating the matrix M_ij of values that can be calculated: kappa=100.0*i+1, beta=100.0*j+1
with eta=1.0, alpha=0.5, rho=0.0
Calculating normalization factor for combinations of kappa and beta:
Iterations necessary to calculate normalize(kappa, beta):
  6   7   9   9  11  15  18  22   x   x
 11  12  15  21  24  31  34   x   x   x
  7  15  21  24  29  33  38   x   x   x
  9  44  21  26  31  39   x   x   x   x
 10  15  23  29  34   x   x   x   x   x
 13  15  26  31   x   x   x   x   x   x
  5  15  27   x   x   x   x   x   x   x
  6   x   x   x   x   x   x   x   x   x
  x   x   x   x   x   x   x   x   x   x
  x   x   x   x   x   x   x   x   x   x
>>> logging.getLogger().setLevel('WARNING')
A test to ensure that the vectors gamma1 ... gamma3 are orthonormal
>>> ks = [
...   fb8(0.0,      0.0,      0.0,    20.0, 0.0),
...   fb8(-0.25*np.pi, -0.25*np.pi, 0.0,    20.0, 0.0),
...   fb8(-0.25*np.pi, -0.25*np.pi, 0.0,    20.0, 5.0),
...   fb8(0.0,      0.0,      0.5*np.pi, 10.0, 7.0),
...   fb8(0.0,      0.0,      0.5*np.pi, 0.1,  0.0),
...   fb8(0.0,      0.0,      0.5*np.pi, 0.1,  0.1),
...   fb8(0.0,      0.0,      0.5*np.pi, 0.1,  8.0),
... ]
>>> pdf_values = [
...   3.18309886184,
...   0.00909519370,
...   0.09865564569,
...   0.59668931662,
...   0.08780030026,
...   0.08768344462,
...   0.00063128997
... ]
>>> for k in ks:
...   assert(np.abs(np.sum(k.gamma1 * k.gamma2)) < 1E-14)
...   assert(np.abs(np.sum(k.gamma1 * k.gamma3)) < 1E-14)
...   assert(np.abs(np.sum(k.gamma3 * k.gamma2)) < 1E-14)
...   assert(np.abs(np.sum(k.gamma1 * k.gamma1) - 1.0) < 1E-14)
...   assert(np.abs(np.sum(k.gamma2 * k.gamma2) - 1.0) < 1E-14)
...   assert(np.abs(np.sum(k.gamma3 * k.gamma3) - 1.0) < 1E-14)
A test to ensure that the pdf() and the pdf_max() are calculated
correctly.
>>> from numpy.random import default_rng
>>> rng = default_rng(808)
>>> for k, pdf_value in zip(ks, pdf_values):
...   assert np.abs(k.pdf(np.array([1.0, 0.0, 0.0])) - pdf_value) < 1E-8
...   assert np.abs(k.log_pdf(np.array([1.0, 0.0, 0.0])) - np.log(pdf_value)) < 1E-8
...   num_samples = 100000
...   xs = rng.normal(0, 1,(num_samples, 3))
...   xs = np.divide(xs, np.reshape(norm(xs, 1), (num_samples, 1)))
...   values = k.pdf(xs, normalize=False)
...   fmax = k.pdf_max(normalize=False)
...   assert np.all(values <= fmax)
...   assert np.any(values > fmax*0.999)
...   values = k.pdf(xs)
...   fmax = k.pdf_max()
...   assert np.all(values <= fmax)
...   assert np.any(values > fmax*0.999)
These are tests to ensure that the coordinate transformations are done correctly
that the functions that generate instances of FB8Distribution are consistent and
that the derivatives are calculated correctly. In addition some more orthogonality
testing is done.
>>> from distribution import *
>>> def test_orth(k):
...   # a bit more orthonormality testing for good measure
...   assert(np.abs(np.sum(k.gamma1 * k.gamma2)) < 1E-14)
...   assert(np.abs(np.sum(k.gamma1 * k.gamma3)) < 1E-14)
...   assert(np.abs(np.sum(k.gamma3 * k.gamma2)) < 1E-14)
...   assert(np.abs(np.sum(k.gamma1 * k.gamma1) - 1.0) < 1E-14)
...   assert(np.abs(np.sum(k.gamma2 * k.gamma2) - 1.0) < 1E-14)
...   assert(np.abs(np.sum(k.gamma3 * k.gamma3) - 1.0) < 1E-14)
...
>>> # generating some specific boundary values and some random values
>>> from functools import partial
>>> upi, u2pi = partial(rng.uniform, 0, np.pi), partial(rng.uniform, -np.pi, 2*np.pi)
>>> thetas, phis, psis = list(upi(925)), list(u2pi(925)), list(u2pi(925))
>>> for a in (0.0, 0.5*np.pi, np.pi):
...   for b in (-np.pi, -0.5*np.pi, 0, 0.5*np.pi, np.pi):
...     for c in (-np.pi, -0.5*np.pi, 0, 0.5*np.pi, np.pi):
...       thetas.append(a)
...       phis.append(b)
...       psis.append(c)
...
>>> # testing consintency of angles (specifically kent())
>>> for theta, phi, psi in zip(thetas, phis, psis):
...   k = fb8(theta, phi, psi, 1.0, 1.0)
...   assert np.abs(theta - k.theta) < 1E-12
...   a = np.abs(phi - k.phi)
...   b = np.abs(psi - k.psi)
...   if theta != 0 and theta != np.pi:
...     assert a < 1E-12 or np.abs(a-2*np.pi) < 1E-12
...     assert b < 1E-12 or np.abs(b-2*np.pi) < 1E-12
...   test_orth(k)
...
>>> # testing consistency of gammas and consistency of back and forth
>>> # calculations between gammas and angles (specifically fb82(), fb83() and fb84())
>>> kappas = rng.normal(0, 2, 1000)**2
>>> betas = rng.normal(0, 2, 1000)**2
>>> for theta, phi, psi, kappa, beta in zip(thetas, phis, psis, kappas, betas):
...   gamma1, gamma2, gamma3 = FB8Distribution.spherical_coordinates_to_gammas(theta, phi, psi)
...   theta, phi, psi = FB8Distribution.gammas_to_spherical_coordinates(gamma1, gamma2)
...   gamma1a, gamma2a, gamma3a = FB8Distribution.spherical_coordinates_to_gammas(theta, phi, psi)
...   assert np.all(np.abs(gamma1a-gamma1) < 1E-12)
...   assert np.all(np.abs(gamma2a-gamma2) < 1E-12)
...   assert np.all(np.abs(gamma3a-gamma3) < 1E-12)
...   k2 = fb82(gamma1, gamma2, gamma3, kappa, beta)
...   assert np.all(np.abs(gamma1 - k2.gamma1) < 1E-12)
...   assert np.all(np.abs(gamma2 - k2.gamma2) < 1E-12)
...   assert np.all(np.abs(gamma3 - k2.gamma3) < 1E-12)
...   A = gamma1*kappa
...   B = gamma2*beta
...   k3 = fb83(A, B)
...   assert np.all(np.abs(gamma1 - k3.gamma1) < 1E-12)
...   assert np.all(np.abs(gamma2 - k3.gamma2) < 1E-12)
...   assert np.all(np.abs(gamma3 - k3.gamma3) < 1E-12)
...   test_orth(k)
...   gamma = np.array([
...     [gamma1[0], gamma2[0], gamma3[0]],
...     [gamma1[1], gamma2[1], gamma3[1]],
...     [gamma1[2], gamma2[2], gamma3[2]],
...   ])
...   k4 = fb84(gamma, kappa, beta)
...   assert np.all(k2.gamma1 == k4.gamma1)
...   assert np.all(k2.gamma2 == k4.gamma2)
...   assert np.all(k2.gamma3 == k4.gamma3)
...
>>> # testing special case for B with zero length (fb83())
>>> for theta, phi, psi, kappa, beta in zip(thetas, phis, psis, kappas, betas):
...   gamma1, gamma2, gamma3 = FB8Distribution.spherical_coordinates_to_gammas(theta, phi, psi)
...   A = gamma1*kappa
...   B = gamma2*0.0
...   k = fb83(A, B)
...   assert np.all(np.abs(gamma1 - k.gamma1) < 1E-12)
...   test_orth(k)
>>> # testing property handlers and cache
>>> for k in ks:
...    G = k.Gamma
...    for attr in 'theta phi psi kappa beta eta alpha rho'.split():
...      curr = k.__getattribute__(attr)
...      k.__setattr__(attr, curr*rng.uniform(0,1))
...      test_orth(k)
...      _ = k.rvs()
...      assert len(k._cached_rvs) > 0
...      k.__setattr__(attr, curr)
...      test_orth(k)
>>> for _ in range(10):
...    test_example_mle(seed=_)
>>> assert test_example_mle2(300)
Testing various combinations of kappa and beta for 300 samples.
MSE of ME is higher than 0.65 times the MLE for beta/kappa < 0.3
MSE of ME is higher than MLE for beta/kappa >= 0.3
MSE of ME is five times higher than MLE for beta/kappa > 0.5
"""

    import doctest

    sys.exit(doctest.testmod(optionflags=doctest.ELLIPSIS, raise_on_error=False)[0])


# -*- coding: utf-8 -*-
"""
.. Authors:
    Novimir pablant <npablant@pppl.gov>

A logging module for XICSRT.

For now, the purpose of this module is simply is set default logging options for
interactive use.
"""

from logging import *

def defaultConfig(level=None, long=False, force=False):
    fmt = '{levelname:>5.5s}:{name:>6.6s}: {message}'
    if long:
        fmt = fmt.replace('6.6s', '12.12s')
    if level is None:
        level = INFO
    basicConfig(format=fmt, style='{', force=force)
    getLogger('xicsrt').setLevel(level)

# Setup defaultConfig() when this module is imported, but do not force.
defaultConfig()


# -*- coding: utf-8 -*-
"""
.. Authors
    Novimir A. Pablant <npblant@pppl.gov>

Contains a set of "external" libraries and tools that are included in the
XICSRT source code. These are included here either because they are not
currently available on pipy or because they have been highly modified.

Programming Notes
-----------------

Make sure not to add anything here that is incompatible with the MIT license!
"""



# -*- coding: utf-8 -*-
"""
.. Authors
    Novimir Pablant <npablant@pppl.gov>
    James Kring <jdk0026@tigermail.auburn.edu>
    Yevgeniy Yakusevich <eugenethree@gmail.com>

A set of mathematical function with jax acceleration. Many of these functions
are exact copies or slight modification of the functions in xicsrt_math. Other
function are specific to this module.

Programming Notes
-----------------

These module was developed to support some specific modeling work by N. Pablant
and is not used in any of the built-in xicsrt code. There is no plan to support
jax generally within xicsrt, so I am not really sure of the best way to handle
this module for the moment. Maybe move it into xicsrt_contrib?
"""

import jax.numpy as np

def toarray_1d(a):
    """
    Convert the input to a ndarray with at least 1 dimension.
    This is similar to the numpy function atleast_1d, but has less overhead
    and is jax compatible.
    """
    a = np.asarray(a)
    if a.ndim == 0:
        a = a.reshape(1)
    return a


def vector_angle(a, b):
    """
    Find the angle between two vectors.
    """
    a_mod = np.linalg.norm(a)
    b_mod = np.linalg.norm(b)
    if a.ndim == 2 & b.ndim == 2:
        dot = np.einsum('ij,ik->i', a/a_mod, b/b_mod)
    elif a.ndim == 1 & b.ndim == 1:
        dot = np.dot(a/a_mod, b/b_mod)
    else:
        raise Exception('Input must have 1 or 2 dimensions.')
    angle = np.arccos(dot)
    return angle


def vector_rotate(a, b, theta):
    """
    Rotate vector a around vector b by an angle theta (radians)

    Programming Notes:
      u: parallel projection of a on b_hat.
      v: perpendicular projection of a on b_hat.
      w: a vector perpendicular to both a and b.
    """

    if a.ndim == 2:
        b_hat = b / np.linalg.norm(b)
        dot = np.einsum('ij,j->i', a, b_hat)
        u = np.einsum('i,j->ij', dot, b_hat)
        v = a - u
        w = np.cross(b_hat, v)
        c = u + v * np.cos(theta) + w * np.sin(theta)
    elif a.ndim == 1:
        b_hat = b / np.linalg.norm(b)
        u = b_hat * np.dot(a, b_hat)
        v = a - u
        w = np.cross(b_hat, v)
        c = u + v * np.cos(theta) + w * np.sin(theta)
    else:
        raise Exception('Input array must be 1d (vector) or 2d (array of vectors)')
    return c


def sinusoidal_spiral(phi, b, r0, theta0):
    r = r0 * (np.sin(theta0 + (b-1)*phi)/np.sin(theta0))**(1/(b-1))
    return r


def point_to_external(point_local, orientation, origin):
    return vector_to_external(point_local, orientation) + origin


def point_to_local(point_external, orientation, origin):
    return vector_to_local(point_external - origin, orientation)


def vector_to_external(vector, orientation):
    if vector.ndim == 2:
        vector = np.einsum('ij,ki->kj', orientation, vector)
    elif vector.ndim == 1:
        vector = np.einsum('ij,i->j', orientation, vector)
    else:
        raise Exception('vector.ndim must be 1 or 2')
    return vector


def vector_to_local(vector, orientation):
    if vector.ndim == 2:
        vector = np.einsum('ji,ki->kj', orientation, vector)
    elif vector.ndim == 1:
        vector = np.einsum('ji,i->j', orientation, vector)
    else:
        raise Exception('vector.ndim must be 1 or 2')
    return vector